#!/usr/bin/env python
#
# Copyright (c) 2026 Authors and contributors
# (see the AUTHORS.rst file for the full list of names)
#
# Released under the GNU Public Licence, v3 or any higher version
# SPDX-License-Identifier: GPL-3.0-or-later
r"""Module for computing spherical dipolar order parameters."""

import logging

import MDAnalysis as mda

from ..core import ProfileSphereBase
from ..lib.util import render_docs, unit_vectors_sphere
from ..lib.weights import diporder_weights

logger = logging.getLogger(__name__)


@render_docs
class DiporderSphere(ProfileSphereBase):
    r"""Spherical dipolar order parameters.

    ${DIPORDER_DESCRIPTION}

    ${CORRELATION_INFO_RADIAL}

    Parameters
    ----------
    ${ATOMGROUP_PARAMETER}
    ${ORDER_PARAMETER_PARAMETER}
    ${PROFILE_SPHERE_CLASS_PARAMETERS}
    ${OUTPUT_PARAMETER}

    Attributes
    ----------
    ${PROFILE_SPHERE_CLASS_ATTRIBUTES}

    """

    def __init__(
        self,
        atomgroup: mda.AtomGroup,
        order_parameter: str = "P0",
        rmin: float = 0,
        rmax: float | None = None,
        bin_width: float = 1,
        bin_method: str = "com",
        grouping: str = "residues",
        refgroup: mda.AtomGroup | None = None,
        unwrap: bool = True,
        pack: bool = True,
        jitter: float = 0.0,
        concfreq: int = 0,
        output: str = "diporder_sphere.dat",
    ) -> None:
        normalization = "volume" if order_parameter == "P0" else "number"

        def get_unit_vectors(atomgroup: mda.AtomGroup, grouping: str):
            return unit_vectors_sphere(
                atomgroup=atomgroup, grouping=grouping, bin_method=bin_method
            )

        super().__init__(
            atomgroup=atomgroup,
            unwrap=unwrap,
            pack=pack,
            jitter=jitter,
            refgroup=refgroup,
            concfreq=concfreq,
            rmin=rmin,
            rmax=rmax,
            bin_width=bin_width,
            grouping=grouping,
            bin_method=bin_method,
            output=output,
            weighting_function=diporder_weights,
            weighting_function_kwargs={
                "order_parameter": order_parameter,
                "get_unit_vectors": get_unit_vectors,
            },
            normalization=normalization,
        )

    def _prepare(self):
        logger.info("Analysis of the spherical dipolar order parameters.")
        super()._prepare()



#!/usr/bin/env python
#
# Copyright (c) 2026 Authors and contributors
# (see the AUTHORS.rst file for the full list of names)
#
# Released under the GNU Public Licence, v3 or any higher version
# SPDX-License-Identifier: GPL-3.0-or-later
"""Module for computing planar dielectric profiles."""

import logging

import MDAnalysis as mda
import numpy as np
import scipy.constants

from ..core import PlanarBase
from ..lib.math import symmetrize
from ..lib.util import charge_neutral, citation_reminder, get_compound, render_docs

logger = logging.getLogger(__name__)


@render_docs
@charge_neutral(filter="error")
class DielectricPlanar(PlanarBase):
    r"""Planar dielectric profiles.

    Computes the parallel :math:`\varepsilon_\parallel(z)` and inverse perpendicular
    (:math:`\varepsilon_\perp^{-1}(r)`) components of the planar dielectric tensor
    :math:`\varepsilon`. The components are binned along the cartesian :math:`z`
    direction yielding the component normal to the surface and defined by the ``dim``
    parameter.

    For usage please refer to the
    :ref:`sphx_glr_generated_examples_dielectrics_dielectric-profiles.py` example and
    for details on the theory see :ref:`dielectric-explanations`.

    For correlation analysis, the norm of the parallel total dipole moment is used.
    ${CORRELATION_INFO}

    Also, please read and cite
    :footcite:t:`schlaichWaterDielectricEffects2016` and Refs.
    :footcite:p:`locheUniversalNonuniversalAspects2020`,
    :footcite:p:`bonthuisProfileStaticPermittivity2012`.

    Parameters
    ----------
    ${ATOMGROUP_PARAMETER}
    ${TEMPERATURE_PARAMETER}
    vcutwidth : float
        Spacing of virtual cuts (bins) along the parallel directions.
    is_3d : bool
        Use 3d-periodic boundary conditions, i.e., include the dipole correction for
        the interaction between periodic images
        :footcite:p:`sternCalculationDielectricPermittivity2003`.
    ${PLANAR_CLASS_PARAMETERS}
    ${SYM_PARAMETER}
    ${BASE_CLASS_PARAMETERS}
    ${OUTPUT_PREFIX_PARAMETER}

    Attributes
    ----------
    ${PLANAR_CLASS_ATTRIBUTES}
    results.eps_par : numpy.ndarray
        Reduced parallel dielectric profile
        :math:`(\varepsilon_\parallel(z) - 1)` of the selected AtomGroup
    results.deps_par : numpy.ndarray
        Uncertainty of parallel dielectric profile
    results.eps_par_self : numpy.ndarray
        Reduced self contribution of parallel dielectric profile
        :math:`(\varepsilon_{\parallel,\mathrm{self}}(z) - 1)`
    results.eps_par_coll : numpy.ndarray
        Reduced collective contribution of parallel dielectric profile
        :math:`(\varepsilon_{\parallel,\mathrm{coll}}(z) - 1)`
    results.eps_perp : numpy.ndarray
        Reduced inverse perpendicular dielectric profile
        :math:`(\varepsilon^{-1}_\perp(z) - 1)`
    results.deps_perp : numpy.ndarray
        Uncertainty of inverse perpendicular dielectric profile
    results.eps_perp_self : numpy.ndarray
        Reduced self contribution of the inverse perpendicular dielectric
        profile :math:`(\varepsilon^{-1}_{\perp,\mathrm{self}}(z) - 1)`
    results.eps_perp_coll : numpy.ndarray
        Reduced collective contribution of the inverse perpendicular dielectric profile
        :math:`(\varepsilon^{-1}_{\perp,\mathrm{coll}}(z) - 1)`


    """

    def __init__(
        self,
        atomgroup: mda.AtomGroup,
        temperature: float = 300,
        vcutwidth: float = 0.1,
        is_3d: bool = False,
        dim: int = 2,
        zmin: float | None = None,
        zmax: float | None = None,
        bin_width: float = 0.5,
        sym: bool = False,
        refgroup: mda.AtomGroup | None = None,
        unwrap: bool = True,
        pack: bool = True,
        jitter: float = 0.0,
        concfreq: int = 0,
        output_prefix: str = "eps",
    ) -> None:
        self._locals = locals()
        wrap_compound = get_compound(atomgroup)

        if zmin is not None or zmax is not None:
            logger.warning(
                "Setting `zmin` and `zmax` might cut off molecules. This will lead to "
                "severe artifacts in the dielectric profiles."
            )

        super().__init__(
            atomgroup=atomgroup,
            unwrap=unwrap,
            pack=pack,
            refgroup=refgroup,
            jitter=jitter,
            dim=dim,
            zmin=zmin,
            zmax=zmax,
            bin_width=bin_width,
            wrap_compound=wrap_compound,
            concfreq=concfreq,
        )
        self.is_3d = is_3d
        self.sym = sym

        self.temperature = temperature
        self.output_prefix = output_prefix
        self.concfreq = concfreq
        self.vcutwidth = vcutwidth

    def _prepare(self) -> None:
        logger.info(
            "Analysis of the parallel and inverse perpendicular "
            "components of the planar dielectric tensor."
        )
        # Print Alex Schlaich citation
        logger.info(citation_reminder("10.1103/PhysRevLett.117.048001"))

        super()._prepare()

        self.comp = get_compound(self.atomgroup)
        ix = self.atomgroup._get_compound_indices(self.comp)
        _, inverse_ix = np.unique(ix, return_inverse=True)
        self.inverse_ix = inverse_ix

    def _single_frame(self) -> float:
        super()._single_frame()

        # precalculate total polarization of the box
        self._obs.M = np.dot(
            self._universe.atoms.charges, self._universe.atoms.positions
        )

        self._obs.M_perp = self._obs.M[self.dim]
        self._obs.M_perp_2 = self._obs.M[self.dim] ** 2
        self._obs.M_par = self._obs.M[self.odims]

        self._obs.m_par = np.zeros((self.n_bins, 2))
        self._obs.mM_par = np.zeros(self.n_bins)
        self._obs.mm_par = np.zeros(self.n_bins)
        self._obs.cmM_par = np.zeros(self.n_bins)
        self._obs.cM_par = np.zeros((self.n_bins, 2))

        self._obs.m_perp = np.zeros(self.n_bins)
        self._obs.mM_perp = np.zeros(self.n_bins)
        self._obs.mm_perp = np.zeros(self.n_bins)
        self._obs.cmM_perp = np.zeros(self.n_bins)
        self._obs.cM_perp = np.zeros(self.n_bins)

        # Use polarization density (for perpendicular component)
        # ======================================================
        zbins = np.digitize(
            self.atomgroup.atoms.positions[:, self.dim], self._obs.bin_edges[1:-1]
        )

        curQ = np.bincount(
            zbins, weights=self.atomgroup.atoms.charges, minlength=self.n_bins
        )

        self._obs.m_perp = -np.cumsum(curQ / self._obs.bin_area)
        self._obs.mM_perp = self._obs.m_perp * self._obs.M_perp
        self._obs.mm_perp = self._obs.m_perp**2 * self._obs.bin_volume
        self._obs.cmM_perp = self._obs.m_perp * (
            self._obs.M_perp - self._obs.m_perp * self._obs.bin_volume
        )

        self._obs.cM_perp = self._obs.M_perp - self._obs.m_perp * self._obs.bin_volume

        # Use virtual cutting method (for parallel component)
        # ===================================================
        # Move all z-positions to 'center of charge' such that we avoid monopoles in
        # z-direction (compare Eq. 33 in Bonthuis 2012; we only want to cut in x/y
        # direction)
        testpos = self.atomgroup.center(
            weights=np.abs(self.atomgroup.charges), compound=self.comp
        )[self.inverse_ix, self.dim]

        # Average parallel directions
        for j, direction in enumerate(self.odims):
            # At this point we should not use the wrap, which causes unphysical
            # dipoles at the borders
            Lx = self._ts.dimensions[direction]
            Ax = self._ts.dimensions[self.odims[1 - j]] * self._obs.bin_width

            vbinsx = np.ceil(Lx / self.vcutwidth).astype(int)
            x_bin_edges = (np.arange(vbinsx)) * (Lx / vbinsx)

            zpos = np.digitize(testpos, self._obs.bin_edges[1:-1])
            xbins = np.digitize(
                self.atomgroup.atoms.positions[:, direction], x_bin_edges[1:]
            )

            curQx = np.bincount(
                zpos + self.n_bins * xbins,
                weights=self.atomgroup.charges,
                minlength=vbinsx * self.n_bins,
            ).reshape(vbinsx, self.n_bins)

            # integral over x, so uniself._ts of area
            self._obs.m_par[:, j] = -np.cumsum(curQx / Ax, axis=0).mean(axis=0)

        # Can not use array for operations below, without extensive reshaping of
        # each array... Therefore, take first element only since the volume of each
        # bin is the same in planar geometry.
        bin_volume = self._obs.bin_volume[0]

        self._obs.mM_par = np.dot(self._obs.m_par, self._obs.M_par)
        self._obs.mm_par = (self._obs.m_par * self._obs.m_par).sum(axis=1) * bin_volume
        self._obs.cmM_par = (
            self._obs.m_par * (self._obs.M_par - self._obs.m_par * bin_volume)
        ).sum(axis=1)
        self._obs.cM_par = self._obs.M_par - self._obs.m_par * bin_volume

        # Save norm of the total parallel dipole moment for correlation analysis.
        return np.linalg.norm(self._obs.M_par)

    def _conclude(self) -> None:
        super()._conclude()

        self._pref = 1 / scipy.constants.epsilon_0
        self._pref /= scipy.constants.Boltzmann * self.temperature
        # Convert from ~e^2/m to ~base units
        self._pref /= (
            scipy.constants.angstrom / (scipy.constants.elementary_charge) ** 2
        )

        self.results.V = self.means.bin_volume.sum()

        # Perpendicular component
        # =======================
        cov_perp = self.means.mM_perp - self.means.m_perp * self.means.M_perp

        # Using propagation of uncertainties
        dcov_perp = np.sqrt(
            self.sems.mM_perp**2
            + (self.means.M_perp * self.sems.m_perp) ** 2
            + (self.means.m_perp * self.sems.M_perp) ** 2
        )

        var_perp = self.means.M_perp_2 - self.means.M_perp**2

        cov_perp_self = self.means.mm_perp - (
            self.means.m_perp**2 * self.means.bin_volume[0]
        )
        cov_perp_coll = self.means.cmM_perp - self.means.m_perp * self.means.cM_perp

        if not self.is_3d:
            self.results.eps_perp = -self._pref * cov_perp
            self.results.eps_perp_self = -self._pref * cov_perp_self
            self.results.eps_perp_coll = -self._pref * cov_perp_coll
            self.results.deps_perp = self._pref * dcov_perp

        else:
            self.results.eps_perp = -cov_perp / (
                self._pref**-1 + var_perp / self.results.V
            )
            self.results.deps_perp = self._pref * dcov_perp

            self.results.eps_perp_self = (-self._pref * cov_perp_self) / (
                1 + self._pref / self.results.V * var_perp
            )
            self.results.eps_perp_coll = (-self._pref * cov_perp_coll) / (
                1 + self._pref / self.results.V * var_perp
            )

        # Parallel component
        # ==================
        cov_par = np.zeros(self.n_bins)
        dcov_par = np.zeros(self.n_bins)
        cov_par_self = np.zeros(self.n_bins)
        cov_par_coll = np.zeros(self.n_bins)

        cov_par = 0.5 * (
            self.means.mM_par - np.dot(self.means.m_par, self.means.M_par.T)
        )

        # Using propagation of uncertainties
        dcov_par = 0.5 * np.sqrt(
            self.sems.mM_par**2
            + np.dot(self.sems.m_par**2, (self.means.M_par**2).T)
            + np.dot(self.means.m_par**2, (self.sems.M_par**2).T)
        )

        cov_par_self = 0.5 * (
            self.means.mm_par - np.dot(self.means.m_par, self.means.m_par.sum(axis=0))
        )
        cov_par_coll = 0.5 * (
            self.means.cmM_par - (self.means.m_par * self.means.cM_par).sum(axis=1)
        )

        self.results.eps_par = self._pref * cov_par
        self.results.deps_par = self._pref * dcov_par
        self.results.eps_par_self = self._pref * cov_par_self
        self.results.eps_par_coll = self._pref * cov_par_coll

        if self.sym:
            symmetrize(self.results.eps_perp, axis=0, inplace=True)
            symmetrize(self.results.deps_perp, axis=0, inplace=True)
            symmetrize(self.results.eps_perp_self, axis=0, inplace=True)
            symmetrize(self.results.eps_perp_coll, axis=0, inplace=True)

            symmetrize(self.results.eps_par, axis=0, inplace=True)
            symmetrize(self.results.deps_par, axis=0, inplace=True)
            symmetrize(self.results.eps_par_self, axis=0, inplace=True)
            symmetrize(self.results.eps_par_coll, axis=0, inplace=True)

    @render_docs
    def save(self) -> None:
        """${SAVE_METHOD_PREFIX_DESCRIPTION}"""  # noqa: D415
        columns = ["position [Å]"]
        columns.append("ε^-1_⟂ - 1")
        columns.append("Δε^-1_⟂")
        columns.append("self ε^-1_⟂ - 1")
        columns.append("coll. ε^-1_⟂ - 1")

        outdata_perp = np.vstack(
            [
                self.results.bin_pos,
                self.results.eps_perp,
                self.results.deps_perp,
                self.results.eps_perp_self,
                self.results.eps_perp_coll,
            ]
        ).T

        self.savetxt(
            "{}{}".format(self.output_prefix, "_perp"), outdata_perp, columns=columns
        )

        columns = ["position [Å]"]
        columns.append("ε_∥ - 1")
        columns.append("Δε_∥")
        columns.append("self ε_∥ - 1")
        columns.append("coll ε_∥ - 1")

        outdata_par = np.vstack(
            [
                self.results.bin_pos,
                self.results.eps_par,
                self.results.deps_par,
                self.results.eps_par_self,
                self.results.eps_par_coll,
            ]
        ).T

        self.savetxt(
            "{}{}".format(self.output_prefix, "_par"), outdata_par, columns=columns
        )



#!/usr/bin/env python
#
# Copyright (c) 2026 Authors and contributors
# (see the AUTHORS.rst file for the full list of names)
#
# Released under the GNU Public Licence, v3 or any higher version
# SPDX-License-Identifier: GPL-3.0-or-later
"""Modules init file."""

from .densitycylinder import DensityCylinder
from .densityplanar import DensityPlanar
from .densitysphere import DensitySphere
from .dielectriccylinder import DielectricCylinder
from .dielectricplanar import DielectricPlanar
from .dielectricspectrum import DielectricSpectrum
from .dielectricsphere import DielectricSphere
from .dipoleangle import DipoleAngle
from .dipordercylinder import DiporderCylinder
from .diporderplanar import DiporderPlanar
from .dipordersphere import DiporderSphere
from .diporderstructurefactor import DiporderStructureFactor
from .kineticenergy import KineticEnergy
from .pdfcylinder import PDFCylinder
from .pdfplanar import PDFPlanar
from .rdfdiporder import RDFDiporder
from .saxs import Saxs
from .temperatureplanar import TemperaturePlanar
from .velocitycylinder import VelocityCylinder
from .velocityplanar import VelocityPlanar

__all__ = [
    "DensityCylinder",
    "DensityPlanar",
    "DensitySphere",
    "DielectricCylinder",
    "DielectricPlanar",
    "DielectricSpectrum",
    "DielectricSphere",
    "DipoleAngle",
    "DiporderCylinder",
    "DiporderPlanar",
    "DiporderSphere",
    "DiporderStructureFactor",
    "KineticEnergy",
    "PDFCylinder",
    "PDFPlanar",
    "RDFDiporder",
    "Saxs",
    "TemperaturePlanar",
    "VelocityCylinder",
    "VelocityPlanar",
]



#!/usr/bin/env python
#
# Copyright (c) 2026 Authors and contributors
# (see the AUTHORS.rst file for the full list of names)
#
# Released under the GNU Public Licence, v3 or any higher version
# SPDX-License-Identifier: GPL-3.0-or-later
"""Small helper and utilities functions that don't fit anywhere else."""

import functools
import inspect
import logging
import re
import sys
import warnings
from collections.abc import Callable
from pathlib import Path
from typing import Protocol

import MDAnalysis as mda
import numpy as np
from scipy.signal import find_peaks

from maicos.lib.math import correlation_time

logger = logging.getLogger(__name__)

DOC_REGEX_PATTERN = re.compile(r"\$\{([^\}]+)\}")

DOC_DICT = dict(
    #####################
    # DESCRIPTION SECTION
    #####################
    SAVE_METHOD_DESCRIPTION="Save results of analysis to file specified by ``output``.",
    SAVE_METHOD_PREFIX_DESCRIPTION=r"""Save results of analysis to files specified by
``output_prefix``.""",
    DENSITY_DESCRIPTION_1=r"""Calculations are carried out for ``mass``
:math:`(\rm u \cdot Å^{-3})`, ``number`` :math:`(\rm Å^{-3})`, partial ``charge``
:math:`(\rm e \cdot Å^{-3})` or electron :math:`(\rm e \cdot Å^{-3})` density
profiles """,
    DENSITY_DESCRIPTION_2="""Cell dimensions are allowed to fluctuate in time.

For grouping with respect to ``molecules``, ``residues`` etc., the corresponding
centers (i.e., center of mass), taking into account periodic boundary conditions,
are calculated. For these calculations molecules will be unwrapped/made whole.
Trajectories containing already whole molecules can be run with ``unwrap=False`` to
gain a speedup. For grouping with respect to atoms, the ``unwrap`` option is always
ignored.""",
    PLANAR_DESCRIPTION="""along certain cartesian axes ``[x, y, z]`` of the simulation
cell.""",
    CYLINDRICAL_DESCRIPTION="""along the radial axis in a cylindrical coordinate system,
with principal axis along ``[x, y, z]`` axes of the simulation cell. The origin of the
coordinate system defaults to the box center, but can be set to dynamically follow the
center of mass of a reference group via ``refgroup``""",
    SPHERICAL_DESCRIPTION="""along radial axis in spherical coordinate system. The
origin of the coordinate system defaults to the box center, but can be set to
dynamically follow the center of mass of a reference group via `refgroup`.""",
    DENSITY_PLANAR_DESCRIPTION=r"""${DENSITY_DESCRIPTION_1}
${PLANAR_DESCRIPTION}
${DENSITY_DESCRIPTION_2}""",
    DENSITY_CYLINDER_DESCRIPTION=r"""${DENSITY_DESCRIPTION_1}
${CYLINDRICAL_DESCRIPTION}
${DENSITY_DESCRIPTION_2}""",
    DENSITY_SPHERE_DESCRIPTION=r"""${DENSITY_DESCRIPTION_1}
${SPHERICAL_DESCRIPTION}
${DENSITY_DESCRIPTION_2}""",
    DIPORDER_DESCRIPTION=r"""Calculations include the projected dipole density
:math:`P_0⋅ρ(z)⋅\cos(θ[z])`, the dipole orientation :math:`\cos(θ[z])`, the squared
dipole orientation :math:`\cos²(Θ[z])` and the number density :math:`ρ(z)`.""",
    CORRELATION_INFO=r"""For further information on the correlation analysis please
refer to :class:`AnalysisBase <maicos.core.base.AnalysisBase>` or the
:ref:`general-design` section.""",
    CORRELATION_INFO_PLANAR=r"""For the correlation analysis the central bin
(:math:`N / 2`) of the 0th's group profile is used. ${CORRELATION_INFO}""",
    CORRELATION_INFO_RADIAL="""For the correlation analysis the 0th bin of the 0th's
group profile is used. ${CORRELATION_INFO}""",
    RUN_METHOD_DESCRIPTION="""Iterate over the trajectory.

Parameters
----------
start : int
    start frame of analysis
stop : int
    stop frame of analysis
step : int
    number of frames to skip between each analysed frame
frames : array_like
    array of integers or booleans to slice trajectory; ``frames`` can only be
    used *instead* of ``start``, ``stop``, and ``step``. Setting *both*
    ``frames`` and at least one of ``start``, ``stop``, ``step`` to a
    non-default value will raise a :exc:`ValueError`.
verbose : bool
    Turn on verbosity
progressbar_kwargs : dict
    ProgressBar keywords with custom parameters regarding progress bar position,
    etc; see :class:`MDAnalysis.lib.log.ProgressBar` for full list.

Returns
-------
self : object
    analysis object
""",
    ##########################
    # SINGLE PARAMETER SECTION
    ##########################
    ATOMGROUP_PARAMETER="""atomgroup : MDAnalysis.core.groups.AtomGroup
    A :class:`~MDAnalysis.core.groups.AtomGroup` for which the calculations are
    performed.""",
    WRAP_COMPOUND_PARAMETER="""wrap_compound : str
    The group which will be kept together through the wrap processes. Allowed values
    are: ``"atoms"``, ``"group"``, ``"residues"``, ``"segments"``, ``"molecules"``, or
    ``"fragments"``.""",
    DENS_PARAMETER="""dens : {``"mass"``, ``"number"``, ``"charge"``, ``"electron"``}
    density type to be calculated.""",
    TEMPERATURE_PARAMETER="""temperature : float
    Reference temperature (K)""",
    BIN_WIDTH_PARAMETER="""bin_width : float
    Width of the bins (in Å).""",
    DIM_PARAMETER="""dim : {0, 1, 2}
    Dimension for binning (``x=0``, ``y=1``, ``z=1``).""",
    VDIM_PARAMETER="""vdim : {0, 1, 2}
    Dimension for velocity binning (``x=0``, ``y=1``, ``z=1``).""",
    PDIM_PLANAR_PARAMETER="""pdim : {0, 1, 2}
    direction of the projection""",
    PDIM_RADIAL_PARAMETER="""pdim : {``"r"``, ``"z"``}
    direction of the projection""",
    FLUX_PARAMETER=r"""flux : bool
    Calculate the flux (:math:`[Å^2/\mathrm{ps}]`) instead of the velocity.""",
    GROUPING_PARAMETER="""grouping : {``"atoms"``, ``"residues"``, ``"segments"``, ``"molecules"``, ``"fragments"``}
    Atom grouping for the calculations.

    The possible grouping options are the atom positions (in the case where
    ``grouping="atoms"``) or the center of mass of the specified grouping unit (in the
    case where ``grouping="residues"``, ``"segments"``, ``"molecules"`` or
    ``"fragments"``).""",  # noqa: E501
    OUTPUT_PARAMETER="""output : str
    Output filename.""",
    OUTPUT_PREFIX_PARAMETER="""output_prefix : str
    Prefix for output files.""",
    SYM_PARAMETER="""sym : bool
    Symmetrize the profile. Only works in combination with ``refgroup``.""",
    BIN_METHOD_PARAMETER="""bin_method : {``"com"``, ``"cog"``, ``"coc"``}
    Method for the position binning.

    The possible options are center of mass (``"com"``), center of geometry (``"cog"``),
    and center of charge (``"coc"``).""",
    ORDER_PARAMETER_PARAMETER="""order_parameter : {``"P0"``, ``"cos_theta"``, ``"cos_2_theta"``}
    Order parameter to be calculated:
        - ``"P0"``: total dipole moment projected on an axis
        - ``"cos_theta"``: cosine of the dipole moment with an axis
        - ``"cos_2_theta"``: squred cosine with an axis.""",  # noqa: E501
    ###################################
    # MULTI/COMBINES PARAMETERS SECTION
    ###################################
    BASE_CLASS_PARAMETERS="""refgroup : MDAnalysis.core.groups.AtomGroup
    Reference :class:`~MDAnalysis.core.groups.AtomGroup` used for the calculation. If
    ``refgroup`` is provided, the calculation is performed relative to the center of
    mass of the AtomGroup. If ``refgroup`` is :obj:`None` the calculations are performed
    with respect to the center of the (changing) box.
unwrap : bool
    When :obj:`True`, molecules that are broken due to the periodic boundary conditions
    are made whole.

    If the input contains molecules that are already whole, speed up the calculation by
    disabling unwrap. To do so, use the flag ``-no-unwrap`` when using MAICoS from the
    command line, or use ``unwrap=False`` when using MAICoS from the Python interpreter.

    Note: Molecules containing virtual sites (e.g. TIP4P water models) are not currently
    supported in MDAnalysis. In this case, you need to provide unwrapped trajectory
    files directly, and disable unwrap. Trajectories can be unwrapped, for example,
    using the ``trjconv`` command of GROMACS.
pack : bool
    When :obj:`True`, molecules are put back into the unit cell. This is required
    because MAICoS only takes into account molecules that are inside the unit cell.

    If the input contains molecules that are already packed, speed up the calculation by
    disabling packing with ``pack=False``.
jitter : float
    Magnitude of the random noise to add to the atomic positions.

    A jitter can be used to stabilize the aliasing effects sometimes appearing when
    histogramming data. The jitter value should be about the precision of the
    trajectory. In that case, using jitter will not alter the results of the histogram.
    If ``jitter = 0.0`` (default), the original atomic positions are kept unchanged.

    You can estimate the precision of the positions in your trajectory with
    :func:`maicos.lib.util.trajectory_precision`. Note that if the precision is not the
    same for all frames, the smallest precision should be used.
concfreq : int
    When concfreq (for conclude frequency) is larger than ``0``, the conclude function
    is called and the output files are written every ``concfreq`` frames.""",
    PROFILE_CLASS_PARAMETERS_PRIVATE="""weighting_function : callable
    The function calculating the array weights for the histogram analysis. It must take
    an :py:class:`AtomGroup<MDAnalysis.AtomGroup>` as first argument and a grouping
    (``"atoms"``, ``"residues"``, ``"segments"``, ``"molecules"``, ``"fragments"``) as
    second. Additional parameters can be given as ``weighting_function_kwargs``. The
    function must return a numpy.ndarray with the same length as the number of group
    members.
weighting_function_kwargs : dict
    Additional keyword arguments for ``weighting_function``
normalization : {``"none"``, ``"number"``, ``"volume"``}
    The normalization of the profile performed in every frame. If :obj:`None`, no
    normalization is performed. If `number`, the histogram is divided by the number of
    occurences in each bin. If `volume`, the profile is divided by the volume of each
    bin.""",
    PLANAR_Z_PARAMETERS="""zmin : float
    Minimal coordinate for evaluation (in Å) with respect to the center of mass of
    the refgroup.
    If ``zmin=None``, all coordinates down to the lower cell boundary are taken into
    account.
zmax : float
    Maximal coordinate for evaluation (in Å) with respect to the center of mass of the
    refgroup.
    If ``zmax = None``, all coordinates up to the upper cell boundary are taken into
    account.""",
    PLANAR_CLASS_PARAMETERS="""${DIM_PARAMETER}
${PLANAR_Z_PARAMETERS}
${BIN_WIDTH_PARAMETER}""",
    RADIAL_CLASS_PARAMETERS="""rmin : float
    Minimal radial coordinate relative to the center of mass of the refgroup for
    evaluation (in Å).
rmax : float
    Maximal radial coordinate relative to the center of mass of the refgroup for
    evaluation (in Å).

    If ``rmax=None``, the box extension is taken.""",
    PDF_PARAMETERS="""g1 : MDAnalysis.core.groups.AtomGroup
    First AtomGroup.
g2 : MDAnalysis.core.groups.AtomGroup
    Second AtomGroup.""",
    PROFILE_CLASS_PARAMETERS="""${BIN_METHOD_PARAMETER}
${GROUPING_PARAMETER}""",
    CYLINDER_CLASS_PARAMETERS="""${DIM_PARAMETER}
${PLANAR_Z_PARAMETERS}
${RADIAL_CLASS_PARAMETERS}
${BIN_WIDTH_PARAMETER}""",
    SPHERE_CLASS_PARAMETERS="""${RADIAL_CLASS_PARAMETERS}
${BIN_WIDTH_PARAMETER}""",
    PROFILE_PLANAR_CLASS_PARAMETERS="""
${PLANAR_CLASS_PARAMETERS}
${PROFILE_CLASS_PARAMETERS}
${SYM_PARAMETER}
${BASE_CLASS_PARAMETERS}""",
    PROFILE_CYLINDER_CLASS_PARAMETERS="""${CYLINDER_CLASS_PARAMETERS}
${PROFILE_CLASS_PARAMETERS}
${BASE_CLASS_PARAMETERS}""",
    PROFILE_SPHERE_CLASS_PARAMETERS="""${SPHERE_CLASS_PARAMETERS}
${PROFILE_CLASS_PARAMETERS}
${BASE_CLASS_PARAMETERS}""",
    ###################
    # ATTRIBUTE SECTION
    ###################
    PLANAR_CLASS_ATTRIBUTES="""results.bin_pos : numpy.ndarray
    Bin positions (in Å) ranging from ``zmin`` to ``zmax``.""",
    RADIAL_CLASS_ATTRIBUTES="""results.bin_pos : numpy.ndarray
    Bin positions (in Å) ranging from ``rmin`` to ``rmax``.""",
    PROFILE_CLASS_ATTRIBUTES="""results.profile : numpy.ndarray
    Calculated profile.
results.dprofile : numpy.ndarray
    Estimated profile's uncertainity.""",
    CYLINDER_CLASS_ATTRIBUTES="${RADIAL_CLASS_ATTRIBUTES}",
    SPHERE_CLASS_ATTRIBUTES="${RADIAL_CLASS_ATTRIBUTES}",
    PROFILE_PLANAR_CLASS_ATTRIBUTES="""${PLANAR_CLASS_ATTRIBUTES}
${PROFILE_CLASS_ATTRIBUTES}""",
    PROFILE_CYLINDER_CLASS_ATTRIBUTES="""${RADIAL_CLASS_ATTRIBUTES}
${PROFILE_CLASS_ATTRIBUTES}""",
    PROFILE_SPHERE_CLASS_ATTRIBUTES="""${RADIAL_CLASS_ATTRIBUTES}
${PROFILE_CLASS_ATTRIBUTES}""",
)
"""Dictionary containing the keys and the actual docstring used by :func:`maicos.lib.util.render_docs`.

    :meta hide-value:
"""  # noqa: E501


def _render_docs(func: Callable, doc_dict: dict = DOC_DICT) -> Callable:
    if func.__doc__ is not None:
        while True:
            keys = DOC_REGEX_PATTERN.findall(func.__doc__)
            if not keys:
                break  # Exit the loop if no more patterns are found
            for key in keys:
                func.__doc__ = func.__doc__.replace(f"${{{key}}}", doc_dict[key])
    return func


def render_docs(func: Callable) -> Callable:
    """Replace all template phrases in the functions docstring.

    Keys for the replacement are taken from in :attr:`maicos.lib.util.DOC_DICT`.

    Parameters
    ----------
    func : callable
        The callable (function, class) where the phrase old should be replaced.

    Returns
    -------
    Callable
        callable with replaced phrase

    """
    return _render_docs(func, doc_dict=DOC_DICT)


def correlation_analysis(timeseries: np.ndarray) -> float:
    """Timeseries correlation analysis.

    Analyses a timeseries for correlation and prints a warning if the correlation time
    is larger than the step size.

    Parameters
    ----------
    timeseries : numpy.ndarray
        Array of (possibly) correlated data.

    Returns
    -------
    corrtime: float
        Estimated correlation time of ``timeseries``.

    """
    if np.any(np.isnan(timeseries)):
        # Fail silently if there are NaNs in the timeseries. This is the case if the
        # feature is not implemented for the given analysis. It could also be because of
        # a bug, but that is not our business.
        return -1
    if len(timeseries) <= 4:
        warnings.warn(
            "Your trajectory is too short to estimate a correlation time. Use the "
            "calculated error estimates with caution.",
            stacklevel=2,
        )
        return -1

    corrtime = correlation_time(timeseries)

    if corrtime == -1:
        warnings.warn(
            "Your trajectory does not provide sufficient statistics to estimate a "
            "correlation time. Use the calculated error estimates with caution.",
            stacklevel=2,
        )
    if corrtime > 0.5:
        warnings.warn(
            "Your data seems to be correlated with a correlation time which is "
            f"{corrtime + 1:.2f} times larger than your step size. Consider increasing "
            f"your step size by a factor of {int(np.ceil(2 * corrtime + 1)):d} to get "
            "a reasonable error estimate.",
            stacklevel=2,
        )
    return corrtime


@render_docs
def get_compound(atomgroup: mda.AtomGroup) -> str:
    """Returns the highest order topology attribute.

    The order is "molecules", "fragments", "residues". If the topology contains none of
    those attributes, an AttributeError is raised.

    Parameters
    ----------
    ${ATOMGROUP_PARAMETER}

    Returns
    -------
    str
        Name of the topology attribute.

    Raises
    ------
    AttributeError
        ``atomgroup`` is missing any connection information"

    """
    if hasattr(atomgroup, "molnums"):
        return "molecules"
    if hasattr(atomgroup, "fragments"):
        logger.info("Cannot use 'molecules'. Falling back to 'fragments'")
        return "fragments"
    if hasattr(atomgroup, "residues"):
        logger.info("Cannot use 'fragments'. Falling back to 'residues'")
        return "residues"
    raise AttributeError("Missing any connection information in `atomgroup`.")


def get_cli_input() -> str:
    """Return a proper formatted string of the command line input.

    Returns
    -------
    str
        A string representing the command line input in a proper format.

    """
    program_name = Path(sys.argv[0]).name
    # Add additional quotes for connected arguments.
    arguments = [f'"{arg}"' if " " in arg else arg for arg in sys.argv[1:]]
    return "{} {}".format(program_name, " ".join(arguments))


def atomgroup_header(AtomGroup: mda.AtomGroup) -> str:
    """Return a string containing infos about the AtomGroup.

    Infos include the total number of atoms, the including residues and the number of
    residues. Useful for writing output file headers.

    Parameters
    ----------
    AtomGroup : MDAnalysis.core.groups.AtomGroup
        The AtomGroup object containing the atoms.

    Returns
    -------
    str
        A string containing the AtomGroup information.

    """
    if not hasattr(AtomGroup, "types"):
        logger.warning(
            "AtomGroup does not contain atom types. Not writing AtomGroup information "
            "to output."
        )
        return f"{len(AtomGroup.atoms)} unkown particles"
    unique, unique_counts = np.unique(AtomGroup.types, return_counts=True)
    return " & ".join("{} {}".format(*i) for i in np.vstack([unique, unique_counts]).T)


def charge_neutral(filter: str) -> Callable:
    """Raise a Warning when AtomGroup is not charge neutral.

    Class Decorator to raise an Error/Warning when AtomGroup in an AnalysisBase class is
    not charge neutral. The behaviour of the warning can be controlled with the filter
    attribute. If the AtomGroup's corresponding universe is non-neutral an ValueError is
    raised.

    Parameters
    ----------
    filter : str
        Filter type to control warning filter. Common values are: "error" or "default"
        See :func:`warnings.simplefilter` for more options.

    """

    def inner(original_class):
        def charge_check(function):
            @functools.wraps(function)
            def wrapped(self):
                if not np.allclose(
                    self.atomgroup.total_charge(compound=get_compound(self.atomgroup)),
                    0,
                    atol=1e-4,
                ):
                    with warnings.catch_warnings():
                        warnings.simplefilter(filter)
                        warnings.warn(
                            "At least one AtomGroup has free charges. Analysis for "
                            "systems with free charges could lead to severe "
                            "artifacts!",
                            stacklevel=1,
                        )

                if not np.allclose(
                    self.atomgroup.universe.atoms.total_charge(), 0, atol=1e-4
                ):
                    raise ValueError(
                        "Analysis for non-neutral systems is not supported."
                    )
                return function(self)

            return wrapped

        original_class._prepare = charge_check(original_class._prepare)

        return original_class

    return inner


def unwrap_refgroup(original_class):
    """Class decorator error if ``unwrap = False`` and ``refgroup != None``."""

    def unwrap_check(function):
        @functools.wraps(function)
        def unwrap_check(self):
            if (
                hasattr(self, "unwrap")
                and hasattr(self, "refgroup")
                and not self.unwrap
                and self.refgroup is not None
            ):
                raise ValueError(
                    "Analysis using `unwrap=False` and `refgroup != None` can lead "
                    "to broken molecules and severe errors."
                )
            return function(self)

        return unwrap_check

    original_class._prepare = unwrap_check(original_class._prepare)

    return original_class


def trajectory_precision(
    trajectory: mda.coordinates.base.ReaderBase, dim: int = 2
) -> np.ndarray:
    """Detect the precision of a trajectory.

    Parameters
    ----------
    trajectory : MDAnalysis.coordinates.base.ReaderBase
        Trajectory from which the precision is detected.
    dim : {2, 0, 1}
        Dimension along which the precision is detected.

    Returns
    -------
    precision : numpy.ndarray
        Precision of each frame of the trajectory.

        If the trajectory has a high precision, its resolution will not be detected, and
        a value of 1e-4 is returned.

    """
    # The threshold will limit the precision of the detection. Using a value that is too
    # low will end up costing a lot of memory. 1e-4 is enough to safely detect the
    # resolution of format like XTC
    threshold_bin_width = 1e-4
    precision = np.zeros(trajectory.n_frames)
    # to be done, add range=(0, -1, 1) parameter for ts in
    # trajectory[range[0]:range[1]:range[2]]:
    for ts in trajectory:
        n_bins = int(
            np.ceil(
                (
                    np.max(trajectory.ts.positions[:, dim])
                    - np.min(trajectory.ts.positions[:, dim])
                )
                / threshold_bin_width
            )
        )
        hist1, z = np.histogram(trajectory.ts.positions[:, dim], bins=n_bins)
        (
            hist2,
            bin_edges,
        ) = np.histogram(np.diff(z[np.where(hist1)]), bins=1000, range=(0, 0.1))
        if len(find_peaks(hist2)[0]) == 0 or bin_edges[find_peaks(hist2)[0][0]] <= 5e-4:
            precision[ts.frame] = 1e-4
        else:
            precision[ts.frame] = bin_edges[find_peaks(hist2)[0][0]]
    return precision


DOI_LIST = {
    "10.1103/PhysRevLett.117.048001": "Schlaich, A. et al., Phys. Rev. Lett. 117, "
    "(2016).",
    "10.1021/acs.jpcb.9b09269": "Loche, P. et al., J. Phys. Chem. B 123, (2019).",
    "10.1021/acs.jpca.0c04063": "Carlson, S. et al., J. Phys. Chem. A 124, (2020).",
    "10.1103/PhysRevE.92.032718": "Schaaf, C. et al., Phys. Rev. E 92, (2015).",
}
"""References associated with MAICoS

    :meta hide-value:
"""


def citation_reminder(*dois: str) -> str:
    """Prints citations in order to remind users to give due credit.

    Parameters
    ----------
    dois : list
        dois associated with the method which calls this. Possible dois are registered
        in :attr:`maicos.lib.util.DOI_LIST`.

    Returns
    -------
    cite : str
        formatted citation reminders

    """
    cite = ""
    for doi in dois:
        lines = [
            "If you use this module in your work, please read and cite:",
            DOI_LIST[doi],
            f"doi: {doi}",
        ]

        plus = f"{max([len(i) for i in lines]) * '+'}"
        lines.insert(0, f"\n{plus}")
        lines.append(f"{plus}\n")

        cite += "\n".join(lines)

    return cite


@render_docs
def get_center(atomgroup: mda.AtomGroup, bin_method: str, compound: str) -> np.ndarray:
    """Center attribute for an :class:`MDAnalysis.core.groups.AtomGroup`.

    This function acts as a wrapper for the
    :meth:`MDAnalysis.core.groups.AtomGroup.center` method, providing a more
    user-friendly interface by automatically determining the appropriate weights based
    on the chosen binning method.

    Parameters
    ----------
    ${ATOMGROUP_PARAMETER}
    ${BIN_METHOD_PARAMETER}
    compound : {``"group"``, ``"segments"``, ``"residues"``, ``"molecules"``, ``"fragments"``}
        The compound to be used in the center calculation. For example, ``"residue"``,
        ``"segment"``, etc.

    Returns
    -------
    np.ndarray
        The coordinates of the calculated center.

    Raises
    ------
    ValueError
        If the provided ``bin_method`` is not one of {``"com"``, ``"cog"``, ``"coc"``}.

    """  # noqa: E501
    if bin_method == "cog":
        weights = None
    elif bin_method == "com":
        weights = atomgroup.masses
    elif bin_method == "coc":
        weights = atomgroup.charges.__abs__()
    else:
        raise ValueError(
            f"'{bin_method}' is an unknown binning method. Use 'cog', 'com' or 'coc'."
        )

    return atomgroup.center(weights=weights, compound=compound)


@render_docs
def unit_vectors_planar(
    atomgroup: mda.AtomGroup,  # noqa: ARG001
    grouping: str,  # noqa: ARG001
    pdim: int,
) -> np.ndarray:
    """Calculate unit vectors in planar geometry.

    Parameters
    ----------
    ${ATOMGROUP_PARAMETER}
    ${GROUPING_PARAMETER}
    ${PDIM_PLANAR_PARAMETER}

    Returns
    -------
    numpy.ndarray
        the unit vector

    """
    unit_vectors = np.zeros(3)
    unit_vectors[pdim] += 1

    return unit_vectors


@render_docs
def unit_vectors_cylinder(
    atomgroup: mda.AtomGroup,
    grouping: str,
    bin_method: str,
    dim: int,
    pdim: str,
) -> np.ndarray:
    """Calculate cylindrical unit vectors in cartesian coordinates.

    Parameters
    ----------
    ${ATOMGROUP_PARAMETER}
    ${GROUPING_PARAMETER}
    ${BIN_METHOD_PARAMETER}
    ${DIM_PARAMETER}
    ${PDIM_RADIAL_PARAMETER}

    Returns
    -------
    numpy.ndarray
        Array of the calculated unit vectors with shape (3,) for ``pdim='z'`` and shape
        (3,n) for ``pdim='r'``. The length of ``n`` depends on the grouping.

    """
    # We do NOT transform ``unit_vectors`` into cylindrical coordinates, because all
    # scalar products in ``dipolar_weights`` will be performed cartesian coordinates!
    if pdim == "r":
        unit_vectors = get_center(
            atomgroup=atomgroup, bin_method=bin_method, compound=grouping
        )

        unit_vectors -= atomgroup.universe.dimensions[:3] / 2

        # set z direction to zero. r in cylindrical coordinates contains only x and y.
        unit_vectors[:, dim] = 0
        unit_vectors /= np.linalg.norm(unit_vectors, axis=1)[:, np.newaxis]
    elif pdim == "z":
        unit_vectors = np.zeros(3)
        unit_vectors[dim] += 1
    else:
        raise ValueError(
            f"'{pdim}' is an unknown direction for the projection. Use 'r' or 'z'."
        )

    return unit_vectors


@render_docs
def unit_vectors_sphere(
    atomgroup: mda.AtomGroup, grouping: str, bin_method: str
) -> np.ndarray:
    """Calculate spherical unit vectors in cartesian coordinates.

    Parameters
    ----------
    ${ATOMGROUP_PARAMETER}
    ${GROUPING_PARAMETER}
    ${BIN_METHOD_PARAMETER}

    Returns
    -------
    numpy.ndarray
        Array of the calculated unit vectors with shape (3,n). The length of ``n``
        depends on the grouping.

    """
    # We do NOT transform ``unit_vectors`` into spherical coordinates, because all
    # scalar products in ``dipolar_weights`` will be performed cartesian coordinates!
    unit_vectors = get_center(
        atomgroup=atomgroup, bin_method=bin_method, compound=grouping
    )

    # shift origin to box center and afterwards normalize
    unit_vectors -= atomgroup.universe.dimensions[:3] / 2
    unit_vectors /= np.linalg.norm(unit_vectors, axis=1)[:, np.newaxis]

    return unit_vectors


def maicos_banner(version: str = "", frame_char: str = "-") -> str:
    """Prints ASCII banner resembling the MAICoS Logo with 80 chars width.

    Parameters
    ----------
    version : str
        Version string to add to the banner.
    frame_char : str
        Character used to as framing around the banner.

    Returns
    -------
    banner : str
        formatted banner

    """
    banner = r"""
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
@                  __  __              _____    _____            _____         @
@    ()----()     |  \/  |     /\     |_   _|  / ____|          / ____|        @
@   /  |     \    | \  / |    /  \      | |   | |        ___   | (___          @
@  () ||| |  ()   | |\/| |   / /\ \     | |   | |       / _ \   \___ \         @
@   \ |||||_ /    | |  | |  / ____ \   _| |_  | |____  | (_) |  ____) |        @"""
    if len(version) < 8:
        banner += rf"""
@    ()----()     |_|  |_| /_/    \_\ |_____|  \_____|  \___/  |_____/ {version:^8}@
@                                                                              @
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
"""
    else:
        banner += rf"""
@    ()----()     |_|  |_| /_/    \_\ |_____|  \_____|  \___/  |_____/         @
@                                                                              @
@{version:^78}@
@                                                                              @
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
"""
    return banner.replace("@", frame_char)


class Unit_vector(Protocol):
    """Protocol class for unit vector methods type hints."""

    def __call__(self, atomgroup: mda.AtomGroup, grouping: str) -> np.ndarray:
        """Call for type hints."""
        ...


def get_module_input_str(module_obj):
    """Make a string with all the modules parameters."""
    module_name = module_obj.__class__.__name__
    # We have to check this since only the modules have the _locals attribute,
    # not the base classes. Yet we still want to test output behaviour of the base
    # classes.
    if hasattr(module_obj, "_locals") and hasattr(module_obj, "_run_locals"):
        sig = inspect.getfullargspec(module_obj.__class__)
        sig.args.remove("self")
        strings = []
        for param in sig.args:
            if type(module_obj._locals[param]) is str:
                string = f"{param}='{module_obj._locals[param]}'"
            elif (
                param == "atomgroup"
                or param == "refgroup"
                and module_obj._locals[param] is not None
            ):
                string = f"{param}=<AtomGroup>"
            else:
                string = f"{param}={module_obj._locals[param]}"
            strings.append(string)
        init_signature = ", ".join(strings)

        sig = inspect.getfullargspec(module_obj.run)
        sig.args.remove("self")
        run_signature = ", ".join(
            [
                (
                    f"{param}='{module_obj._run_locals[param]}'"
                    if type(module_obj._run_locals[param]) is str
                    else f"{param}={module_obj._run_locals[param]}"
                )
                for param in sig.args
            ]
        )

        module_input = f"{module_name}({init_signature}).run({run_signature})"
    else:
        module_input = f"{module_name}(*args).run(*args)"

    return module_input



#!/usr/bin/env python
#
# Copyright (c) 2026 Authors and contributors
# (see the AUTHORS.rst file for the full list of names)
#
# Released under the GNU Public Licence, v3 or any higher version
# SPDX-License-Identifier: GPL-3.0-or-later
"""MAICoS: Molecular Analysis of Interfacial and COnfined Systems."""

import warnings

from ._version import __version__  # noqa: F401
from .modules import *  # noqa: F403
from .modules import __all__ as __all__

__authors__ = "MAICoS Developer Team"

# Print maicos DeprecationWarnings
warnings.filterwarnings(action="once", category=DeprecationWarning, module="maicos")



#!/usr/bin/env python
#
# Copyright (c) 2026 Authors and contributors
# (see the AUTHORS.rst file for the full list of names)
#
# Released under the GNU Public Licence, v3 or any higher version
# SPDX-License-Identifier: GPL-3.0-or-later
"""Weight functions used for spatial binned analysis modules."""

import MDAnalysis as mda
import numpy as np
from scipy import constants

from .tables import electron_count
from .util import Unit_vector, render_docs


def _resolve_electron_count(element: str) -> float:
    if element == "CH1":
        return _resolve_electron_count("C") + _resolve_electron_count("H")
    if element == "CH2":
        return _resolve_electron_count("C") + 2 * _resolve_electron_count("H")
    if element == "CH3":
        return _resolve_electron_count("C") + 3 * _resolve_electron_count("H")
    if element == "CH4":
        return _resolve_electron_count("C") + 4 * _resolve_electron_count("H")
    if element == "NH1":
        return _resolve_electron_count("N") + _resolve_electron_count("H")
    if element == "NH2":
        return _resolve_electron_count("N") + 2 * _resolve_electron_count("H")
    if element == "NH3":
        return _resolve_electron_count("N") + 3 * _resolve_electron_count("H")
    try:
        return electron_count[element.title()]
    except KeyError as e:
        raise KeyError(
            f"Element '{e.args[0]}' not found. Known elements are listed in the "
            "`maicos.lib.tables.elements` set."
        ) from e


@render_docs
def density_weights(atomgroup: mda.AtomGroup, grouping: str, dens: str) -> np.ndarray:
    """Weights for density calculations.

    Parameters
    ----------
    ${ATOMGROUP_PARAMETER}
    ${GROUPING_PARAMETER}
    ${DENS_PARAMETER}

    Returns
    -------
    numpy.ndarray
        1D array of calculated weights. The length depends on the grouping.

    Raises
    ------
    ValueError
        if grouping or dens parameter is not supported.

    """
    if grouping not in ["atoms", "residues", "segments", "molecules", "fragments"]:
        raise ValueError(
            f"'{grouping}' grouping is not supported. "
            "Use `atoms`, `residues`, `segments`, `molecules` or `fragments`."
        )

    if dens == "number":
        # There exist no properrty like n_molecules
        if grouping == "molecules":
            numbers = len(np.unique(atomgroup.molnums))
        else:
            numbers = getattr(atomgroup, f"n_{grouping}")
        weights = np.ones(numbers)
    elif dens == "mass":
        if grouping == "atoms":
            weights = atomgroup.masses
        else:
            weights = atomgroup.total_mass(compound=grouping)
    elif dens == "charge":
        if grouping == "atoms":
            weights = atomgroup.charges
        else:
            weights = atomgroup.total_charge(compound=grouping)
    elif dens == "electron":
        weights = np.array([_resolve_electron_count(el) for el in atomgroup.elements])
        if grouping != "atoms":
            weights = atomgroup.accumulate(weights, compound=grouping)
    else:
        raise ValueError(
            f"'{dens}' density type is not supported. Use 'mass', 'number', 'charge' "
            "or 'electron'."
        )

    return weights


@render_docs
def temperature_weights(atomgroup: mda.AtomGroup, grouping: str) -> np.ndarray:
    """Weights for temperature calculations.

    Parameters
    ----------
    ${ATOMGROUP_PARAMETER}
    ${GROUPING_PARAMETER}

    Returns
    -------
    numpy.ndarray
        1D array of calculated weights. The length depends on the grouping.

    Raises
    ------
    NotImplementedError
        Currently only works for ``grouping='atoms'``

    """
    if grouping != "atoms":
        raise NotImplementedError(
            f"Temperature calculations of '{grouping}' is not supported. Use 'atoms' "
            "instead.'"
        )

    # ((1 u * Å^2) / (ps^2)) / Boltzmann constant
    prefac = constants.atomic_mass * 1e4 / constants.Boltzmann
    return (atomgroup.velocities**2).sum(axis=1) * atomgroup.atoms.masses / 2 * prefac


@render_docs
def diporder_weights(
    atomgroup: mda.AtomGroup,
    grouping: str,
    order_parameter: str,
    get_unit_vectors: Unit_vector,
) -> np.ndarray:
    """Weights for general diporder calculations.

    Parameters
    ----------
    ${ATOMGROUP_PARAMETER}
    ${GROUPING_PARAMETER}
    ${ORDER_PARAMETER_PARAMETER}
    get_unit_vectors : Callable
        Callable that returns unit vectors on which the projection is performed.
        Returned unit_vectors can either be of shape (3,) or of shape (n, 3). For a
        shape of (3,) the same unit vector is used for all calculations.

    """
    dipoles = atomgroup.dipole_vector(compound=grouping)

    unit_vectors = get_unit_vectors(atomgroup=atomgroup, grouping=grouping)

    # Extend unit_vectors to be of the same length as of dipoles
    if unit_vectors.shape == (3,):
        np.tile(unit_vectors, len(dipoles)).reshape(len(dipoles), 3)
    elif unit_vectors.shape != (len(dipoles), 3):
        raise ValueError(
            f"Returned unit vectors have shape {unit_vectors.shape}. But only shape "
            f"(3,) or {(len(dipoles), 3)} is allowed."
        )

    if order_parameter == "P0":
        weights = np.sum(dipoles * unit_vectors, axis=1)
    elif order_parameter in ["cos_theta", "cos_2_theta"]:
        weights = np.sum(
            dipoles / np.linalg.norm(dipoles, axis=1)[:, np.newaxis] * unit_vectors,
            axis=1,
        )
        if order_parameter == "cos_2_theta":
            weights *= weights
    else:
        raise ValueError(
            f"'{order_parameter}' not supported. "
            "Use 'P0', 'cos_theta' or 'cos_2_theta'."
        )

    return weights


@render_docs
def velocity_weights(atomgroup: mda.AtomGroup, grouping: str, vdim: int) -> np.ndarray:
    """Weights for velocity calculations.

    The function normalises by the number of compounds.

    Parameters
    ----------
    ${ATOMGROUP_PARAMETER}
    ${GROUPING_PARAMETER}
    ${VDIM_PARAMETER}

    Returns
    -------
    numpy.ndarray
        1D array of calculated weights. The length depends on the grouping.

    """
    atom_vels = atomgroup.velocities[:, vdim]

    if grouping == "atoms":
        vels = atom_vels
    else:
        mass_vels = atomgroup.atoms.accumulate(
            atom_vels * atomgroup.atoms.masses, compound=grouping
        )
        group_mass = atomgroup.atoms.accumulate(
            atomgroup.atoms.masses, compound=grouping
        )
        vels = mass_vels / group_mass

    return vels



#!/usr/bin/env python
#
# Copyright (c) 2026 Authors and contributors
# (see the AUTHORS.rst file for the full list of names)
#
# Released under the GNU Public Licence, v3 or any higher version
# SPDX-License-Identifier: GPL-3.0-or-later
r"""Module for computing 1D cylindrical pair distribution functions."""

import logging

import MDAnalysis as mda
import numpy as np
from MDAnalysis.lib.distances import capped_distance

from ..core import CylinderBase
from ..lib.math import transform_cylinder
from ..lib.util import get_center, get_compound, render_docs

logger = logging.getLogger(__name__)


@render_docs
class PDFCylinder(CylinderBase):
    r"""Shell-wise one-dimensional (cylindrical) pair distribution functions.

    The one-dimensional pair distribution functions :math:`g_{\text{1d}}(\phi)`
    and :math:`g_{\text{1d}}(z)` describes the pair distribution to particles
    which lie on the same cylinder along the angular and axial directions
    respectively. These functions can be used in cylindrical systems that are
    inhomogeneous along radial coordinate, and homogeneous in the angular and
    axial directions. It gives the average number density of ``g2`` as a
    function of angular and axial distances respectively from a ``g1`` atom.
    Then the angular pair distribution function is

    .. math::

            g_{\text{1d}}(\phi) = \left \langle \sum_{i}^{N_{g_1}}
            \sum_{j}^{N_{g2}} \delta(\phi - \phi_{ij}) \delta(R_{ij}) \delta(z_{ij})
            \right \rangle


    And the axial pair distribution function is

    .. math::

            g_{\text{1d}}(z) = \left \langle \sum_{i}^{N_{g_1}}
            \sum_{j}^{N_{g2}} \delta(z - z_{ij}) \delta(R_{ij}) \delta(\phi_{ij})
            \right \rangle

    Even though due to consistency reasons the results are called pair distribution
    functions the output is not unitless. The default output is is in dimension of
    number/volume in :math:`Å^{-3}`. If ``density`` is set to :py:obj:`True`, the
    output is normalised by the density of ``g2``.

    Parameters
    ----------
    ${ATOMGROUP_PARAMETER}
    ${PDF_PARAMETERS}
    bin_width_pdf_z : float
        Binwidth of bins in the histogram of the axial PDF (Å).
    bin_width_pdf_phi : float
        Binwidth of bins in the histogram of the angular PDF (Å).
    drwidth : float
        radial width of a PDF cylindrical shell (Å), and axial or angular (arc) slices.
    dmin: float
        the minimum pairwise distance between ``'g1'`` and ``'g2'`` (Å).
    dmax : float
        the maximum pairwise distance between ``'g1'`` and ``'g2'`` (Å).
    density : bool
        normalise the PDF by the density of ``'g2'`` (:math:`Å^{-3}`).
    origin : numpy.ndarray
        Set origin of the cylindrical coordinate system (x,y,z). If :obj:`None` the
        origin will be set according to the ``refgroup`` parameter.
    ${CYLINDER_CLASS_PARAMETERS}
    ${BIN_METHOD_PARAMETER}
    ${BASE_CLASS_PARAMETERS}
    ${OUTPUT_PARAMETER}

    Attributes
    ----------
    ${CYLINDER_CLASS_ATTRIBUTES}
    results.phi_bins: numpy.ndarray
        Angular distances to which the PDF is calculated with shape (``pdf_nbins``) (Å)
    results.z_bins: numpy.ndarray
        axial distances to which the PDF is calculated with shape (``pdf_nbins``) (Å)
    results.phi_pdf: numpy.ndarray
        Angular PDF with shape (``pdf_nbins``, ``n_bins``) (:math:`\text{Å}^{-3}`)
    results.z_pdf: numpy.ndarray
        Axial PDF with shape (``pdf_nbins``, ``n_bins``) (:math:`\text{Å}^{-3}`)

    """

    def __init__(
        self,
        g1: mda.AtomGroup,
        g2: mda.AtomGroup | None = None,
        bin_width_pdf_z: float = 0.3,
        bin_width_pdf_phi: float = 0.1,
        drwidth: float = 0.1,
        dmin: float | None = None,
        dmax: float | None = None,
        density: bool = False,
        origin: np.ndarray | None = None,
        dim: int = 2,
        zmin: float | None = None,
        zmax: float | None = None,
        rmin: float = 0,
        rmax: float | None = None,
        bin_width: float = 1,
        bin_method: str = "com",
        refgroup: mda.AtomGroup | None = None,
        unwrap: bool = False,
        pack: bool = True,
        jitter: float = 0.0,
        concfreq: int = 0,
        output: str = "pdf.dat",
    ) -> None:
        self.comp_1 = get_compound(g1)
        super().__init__(
            atomgroup=g1,
            refgroup=refgroup,
            unwrap=unwrap,
            pack=pack,
            concfreq=concfreq,
            jitter=jitter,
            dim=dim,
            rmin=rmin,
            rmax=rmax,
            zmin=zmin,
            zmax=zmax,
            bin_width=bin_width,
            wrap_compound=self.comp_1,
        )

        self.g1 = g1
        if g2 is None:
            self.g2 = g1
        else:
            self.g2 = g2

        self.bin_width_pdf_phi = bin_width_pdf_phi
        self.bin_width_pdf_z = bin_width_pdf_z
        self.drwidth = drwidth
        self.bin_width = bin_width
        self.output = output
        self.bin_method = bin_method.lower()

        if origin is not None and origin.shape != (3,):
            raise ValueError(
                f"Origin has length {origin.shape} but only (3,) is allowed."
            )
        self.origin = origin

        self.comp_2 = get_compound(self.g2)
        self.nbins_pdf_phi = 100
        self.nbins_pdf_z = 100

        self.dmin = dmin
        self.dmax = dmax
        self.density = density

    def _prepare(self) -> None:
        super()._prepare()
        logger.info("Analysis of the cylindrical pair distribution function.")

        if self.origin is None:
            self.origin = self.box_center

        if self.dmin is None:
            self.dmin = 0
        if self.dmax is None:
            self.dmax = self.box_center[self.dim]
        else:
            if self.dmax > self.box_center[self.dim]:
                raise ValueError(
                    "Axial range of PDF exceeds half of the box size. "
                    "This will lead to unexpected results."
                )

        if self.bin_width_pdf_z > 0:
            self.nbins_pdf_z = int(
                np.ceil((self.dmax - self.dmin) / self.bin_width_pdf_z)
            )
            self.bin_width_pdf_z = (self.dmax - self.dmin) / self.nbins_pdf_z
        else:
            raise ValueError("PDF bin_width must be a positive number.")
        if self.bin_width_pdf_phi > 0:
            self.nbins_pdf_phi = int(np.ceil(np.pi / self.bin_width_pdf_phi))
            self.bin_width_pdf_phi = np.pi / self.nbins_pdf_phi
        else:
            raise ValueError("PDF bin_width must be a positive number.")

        if self.bin_method not in ["cog", "com", "coc"]:
            raise ValueError(
                f"{self.bin_method} is an unknown binning method. Use `cog`, `com` or "
                "`coc`."
            )

        logger.info(
            f"Using {self.nbins_pdf_phi} pdf bins in phi direction and "
            f"{self.nbins_pdf_z} in z direction."
        )

    def _single_frame(self) -> None:
        super()._single_frame()
        self._obs.n_g1 = np.zeros((self.n_bins, 1))
        self._obs.n_g2 = np.zeros((self.n_bins, 1))
        self._obs.count_phi = np.zeros((self.n_bins, self.nbins_pdf_phi))
        self._obs.count_z = np.zeros((self.n_bins, self.nbins_pdf_z))

        # Get the center of each atom in g1 and g2.
        g1_bin_positions = get_center(
            atomgroup=self.g1, bin_method=self.bin_method, compound=self.comp_1
        )
        g2_bin_positions = get_center(
            atomgroup=self.g2, bin_method=self.bin_method, compound=self.comp_2
        )

        # convert to cylinderical coordinates
        g1_bin_positions_cyl = transform_cylinder(
            g1_bin_positions, origin=self.origin, dim=self.dim
        )
        g2_bin_positions_cyl = transform_cylinder(
            g2_bin_positions, origin=self.origin, dim=self.dim
        )

        # Calculate pdf per bin by averaging over all atoms in one bin.
        for r_bin in range(0, self.n_bins):
            # Get all atoms in a bin.
            g1_in_rbin_positions = g1_bin_positions_cyl[
                np.logical_and(
                    g1_bin_positions_cyl[:, 0] >= self._obs.bin_edges[r_bin],
                    g1_bin_positions_cyl[:, 0] < self._obs.bin_edges[r_bin + 1],
                )
            ]

            g2_in_rbin_positions = g2_bin_positions_cyl[
                np.logical_and(
                    g2_bin_positions_cyl[:, 0]
                    >= self._obs.bin_edges[r_bin] - self.drwidth,
                    g2_bin_positions_cyl[:, 0]
                    < self._obs.bin_edges[r_bin + 1] + self.drwidth,
                )
            ]

            self._obs.n_g1[r_bin] = len(g1_in_rbin_positions)
            self._obs.n_g2[r_bin] = len(g2_in_rbin_positions)

            # Below we abuse the 3D `capped_distance` search for do a distance search in
            # 1D distance by setting the other positions as well as the box size in
            # these directions to 0.

            # Filter only those pairs with delta r < dr.
            r_pairs = capped_distance(
                g1_in_rbin_positions * [1, 0, 0],
                g2_in_rbin_positions * [1, 0, 0],
                self.drwidth,
                box=None,
                return_distances=False,
            )

            # Filter only those pairs with delta phi < dphi.
            phi_pairs = capped_distance(
                g1_in_rbin_positions * [0, 1, 0],
                g2_in_rbin_positions * [0, 1, 0],
                # define: r dphi = drwidth
                # therefore: dphi = drwidth / r
                self.drwidth / self._obs.bin_pos[r_bin],
                box=[0, 2 * np.pi, 0, 90, 90, 90],
                return_distances=False,
            )

            # Filter only those pairs with delta z < dz.
            z_pairs = capped_distance(
                g1_in_rbin_positions * [0, 0, 1],
                g2_in_rbin_positions * [0, 0, 1],
                self.drwidth,  # define: dz = drwidth
                box=[0, 0, self._universe.dimensions[self.dim], 90, 90, 90],
                return_distances=False,
            )

            # Calculate pairwise phi distances between g1 and g2.
            phi_dist_pairs, phi_distances = capped_distance(
                g1_in_rbin_positions * [0, 1, 0],
                g2_in_rbin_positions * [0, 1, 0],
                np.pi,  # maximum phi distance is pi
                box=[
                    0,
                    2 * np.pi,
                    0,
                    90,
                    90,
                    90,
                ],  # minimum image convention in phi direction (0, 2pi)
            )

            # Calculate pairwise z distances between g1 and g2.
            z_dist_pairs, z_distances = capped_distance(
                g1_in_rbin_positions * [0, 0, 1],
                g2_in_rbin_positions * [0, 0, 1],
                self.dmax,
                box=[
                    0,
                    0,
                    self._universe.dimensions[self.dim],
                    90,
                    90,
                    90,
                ],  # minimum image convention in z direction (0, boxsize)
            )

            # Map pairs (i, j) to a number i+N*j (so we can use np.isin).
            r_pairs_encode = r_pairs[:, 0] + self._obs.n_g2[r_bin] * r_pairs[:, 1]
            phi_pairs_encode = phi_pairs[:, 0] + self._obs.n_g2[r_bin] * phi_pairs[:, 1]
            z_pairs_encode = z_pairs[:, 0] + self._obs.n_g2[r_bin] * z_pairs[:, 1]
            phi_dist_pairs_encode = (
                phi_dist_pairs[:, 0] + self._obs.n_g2[r_bin] * phi_dist_pairs[:, 1]
            )
            z_dist_pairs_encode = (
                z_dist_pairs[:, 0] + self._obs.n_g2[r_bin] * z_dist_pairs[:, 1]
            )

            # Filter pairs that are in the same dr bin and dz bin.
            mask_in_dr_and_dz = np.isin(
                phi_dist_pairs_encode, r_pairs_encode
            ) * np.isin(phi_dist_pairs_encode, z_pairs_encode)

            # Filter pairs that are in the same dr bin and dphi bin.
            mask_in_dr_and_dphi = np.isin(
                z_dist_pairs_encode, r_pairs_encode
            ) * np.isin(z_dist_pairs_encode, phi_pairs_encode)

            mask_same_atom = phi_distances > 0
            relevant_phi_distances = phi_distances[mask_in_dr_and_dz * mask_same_atom]

            mask_same_atom = z_distances > 0
            relevant_z_distances = z_distances[mask_in_dr_and_dphi * mask_same_atom]

            # Histogram the pairwise distances.
            self._obs.count_phi[r_bin] = np.histogram(
                relevant_phi_distances, bins=self.nbins_pdf_phi, range=(0, np.pi)
            )[0]

            self._obs.count_z[r_bin] = np.histogram(
                relevant_z_distances,
                bins=self.nbins_pdf_z,
                range=(self.dmin, self.dmax),
            )[0]

    def _conclude(self) -> None:
        super()._conclude()

        # Calculate the density of g2.
        g2_density = self.means.n_g2 / self.means.bin_volume if self.density else 1

        # Normalising volume for the angular PDF. This is 2(R*dR*2dz*2dphi),
        # where R is the radius of the bin center, dR is the width of the pdf bin.
        phi_norm = (
            np.array(
                [
                    2
                    * (self.means.bin_edges[1:] + self.means.bin_edges[:-1])
                    / 2
                    * self.bin_width_pdf_phi
                    * 2
                    * self.drwidth
                    * 2
                    * self.drwidth
                ]
            ).T
            * g2_density
        )

        # Normalising volume for the axial PDF. This is 2(dZ*2dr*2dz),
        # where dZ is the width of the pdf bin.
        z_norm = (
            2 * self.bin_width_pdf_z * 2 * self.drwidth * 2 * self.drwidth * g2_density
        )

        # Normalise pdf using the normalisation factor.
        with np.errstate(invalid="ignore", divide="ignore"):
            pdf_phi = self.means.count_phi / self.means.n_g1 / phi_norm
        self.results.pdf_phi = np.nan_to_num(pdf_phi, nan=0, posinf=0, neginf=0)

        with np.errstate(invalid="ignore", divide="ignore"):
            pdf_z = self.means.count_z / self.means.n_g1 / z_norm
        self.results.pdf_z = np.nan_to_num(pdf_z, nan=0, posinf=0, neginf=0)

        # Calculate the bin centers.
        edges_phi = np.histogram([-1], bins=self.nbins_pdf_phi, range=(0, np.pi))[1]
        edges_z = np.histogram(
            [-1], bins=self.nbins_pdf_z, range=(self.dmin, self.dmax)
        )[1]
        self.results.bins_phi = 0.5 * (edges_phi[1:] + edges_phi[:-1])
        self.results.bins_z = 0.5 * (edges_z[1:] + edges_z[:-1])

    @render_docs
    def save(self) -> None:
        """${SAVE_METHOD_DESCRIPTION}"""  # noqa: D415
        columns = ["r [Å]"]
        for r in self.results.bin_pos:
            columns.append(f"pdf at {r:.2f} Å [Å^-3]")

        self.savetxt(
            "phi_" + self.output,
            np.hstack([self.results.bins_phi[:, np.newaxis], self.results.pdf_phi.T]),
            columns=columns,
        )

        self.savetxt(
            "z_" + self.output,
            np.hstack([self.results.bins_z[:, np.newaxis], self.results.pdf_z.T]),
            columns=columns,
        )



#!/usr/bin/env python
#
# Copyright (c) 2026 Authors and contributors
# (see the AUTHORS.rst file for the full list of names)
#
# Released under the GNU Public Licence, v3 or any higher version
# SPDX-License-Identifier: GPL-3.0-or-later
"""Module for computing cylindrical dielectric profiles."""

import logging

import MDAnalysis as mda
import numpy as np
import scipy.constants

from ..core import CylinderBase
from ..lib.util import charge_neutral, citation_reminder, get_compound, render_docs

logger = logging.getLogger(__name__)


@render_docs
@charge_neutral(filter="error")
class DielectricCylinder(CylinderBase):
    r"""Cylindrical dielectric profiles.

    Computes the axial :math:`\varepsilon_z(r)` and inverse radial
    :math:`\varepsilon_r^{-1}(r)` components of the cylindrical dielectric tensor
    :math:`\varepsilon`. The components are binned along the radial direction of the
    cylinder. The :math:`z`-axis of the cylinder is pointing in the direction given by
    the ``dim`` parameter. The center of the cylinder is either located at the center of
    the simulation box (default) or at the center of mass of the ``refgroup``, if
    provided.

    For usage please refer to the
    :ref:`sphx_glr_generated_examples_dielectrics_dielectric-profiles.py` example and
    for details on the theory see :ref:`dielectric-explanations`.

    For correlation analysis, the component along the :math:`z`-axis is used.
    ${CORRELATION_INFO}

    Also, please read and cite :footcite:p:`locheGiantaxialDielectric2019`.

    Parameters
    ----------
    ${ATOMGROUP_PARAMETER}
    ${TEMPERATURE_PARAMETER}
    vcutwidth : float
        Spacing of virtual cuts (bins) along the parallel directions.
    single : bool
        For a single chain of molecules the average of :math:`M` is zero. This flag sets
        :math:`\langle M \rangle = 0`.
    ${CYLINDER_CLASS_PARAMETERS}
    ${BASE_CLASS_PARAMETERS}
    ${OUTPUT_PREFIX_PARAMETER}

    Attributes
    ----------
    ${CYLINDER_CLASS_ATTRIBUTES}
    results.eps_z : numpy.ndarray
        Reduced axial dielectric profile :math:`(\varepsilon_z(r) - 1)` of the
        selected atomgroup
    results.deps_z : numpy.ndarray
        Estimated uncertainty of axial dielectric profile
    results.eps_r : numpy.ndarray
        Reduced inverse radial dielectric profile
        :math:`(\varepsilon^{-1}_r(r) - 1)`
    results.deps_r : numpy.ndarray
        Estimated uncertainty of inverse radial dielectric profile

    """

    def __init__(
        self,
        atomgroup: mda.AtomGroup,
        temperature: float = 300,
        vcutwidth: float = 0.1,
        single: bool = False,
        dim: int = 2,
        zmin: float | None = None,
        zmax: float | None = None,
        rmin: float = 0,
        rmax: float | None = None,
        bin_width: float = 0.1,
        refgroup: mda.AtomGroup | None = None,
        unwrap: bool = True,
        pack: bool = True,
        jitter: float = 0.0,
        concfreq: int = 0,
        output_prefix: str = "eps_cyl",
    ) -> None:
        self._locals = locals()
        self.comp = get_compound(atomgroup)
        ix = atomgroup._get_compound_indices(self.comp)
        _, self.inverse_ix = np.unique(ix, return_inverse=True)

        if zmin is not None or zmax is not None or rmin != 0 or rmax is not None:
            logger.warning(
                "Setting `rmin` and `rmax` (as well as `zmin` and `zmax`) might cut "
                "off molecules. This will lead to severe artifacts in the dielectric "
                "profiles."
            )

        super().__init__(
            atomgroup,
            unwrap=unwrap,
            pack=pack,
            refgroup=refgroup,
            jitter=jitter,
            concfreq=concfreq,
            dim=dim,
            zmin=zmin,
            zmax=zmax,
            bin_width=bin_width,
            rmin=rmin,
            rmax=rmax,
            wrap_compound=self.comp,
        )
        self.output_prefix = output_prefix
        self.temperature = temperature
        self.single = single
        self.vcutwidth = vcutwidth

    def _prepare(self) -> None:
        logger.info(
            "Analysis of the axial and inverse radial "
            "components of the cylindrical dielectric tensor."
        )
        # Print Philip Loche citation
        logger.info(citation_reminder("10.1021/acs.jpcb.9b09269"))

        super()._prepare()

    def _single_frame(self) -> float:
        super()._single_frame()

        # Precalculate the bins each atom belongs to.
        rbins = np.digitize(self.pos_cyl[:, 0], self._obs.bin_edges[1:-1])

        # Calculate the charge per bin for the selected atomgroup.
        curQ_r = np.bincount(
            rbins[self.atomgroup.ix],
            weights=self.atomgroup.charges,
            minlength=self.n_bins,
        )

        # In literature, the charge density is integrated along the radial direction to
        # get the dipole moment density. We can rewrite the integral by identifying:
        # q(a) = 2π * L * int_0^a ρ(r) * r dr,
        # where q(a) is the charge enclosed within a cylinder of radius a and length L.
        # This allows us to avoid numerical errors.
        self._obs.m_r = -np.cumsum(curQ_r) / 2 / np.pi / self._obs.L / self._obs.bin_pos

        # Same as above, but for the whole system.
        curQ_r_tot = np.bincount(
            rbins, weights=self._universe.atoms.charges, minlength=self.n_bins
        )

        self._obs.m_r_tot = (
            -np.cumsum(curQ_r_tot) / 2 / np.pi / self._obs.L / self._obs.bin_pos
        )

        # Note that M_r is not really the total system dipole moment in radial
        # direction, but it keeps the nomenclature consistent across all of the
        # dielectric modules.
        self._obs.M_r = np.sum(self._obs.m_r_tot * self._obs.bin_width)
        self._obs.mM_r = self._obs.m_r * self._obs.M_r

        # Use virtual cutting method (for axial component)
        # ========================================================
        # number of virtual cuts ("many")
        nbinsz = np.ceil(self._obs.L / self.vcutwidth).astype(int)

        # Move all r-positions to 'center of charge' such that we avoid monopoles in
        # r-direction. We only want to cut in z direction.
        chargepos = self.pos_cyl[self.atomgroup.ix, 0] * np.abs(self.atomgroup.charges)
        center = self.atomgroup.accumulate(
            chargepos, compound=self.comp
        ) / self.atomgroup.accumulate(
            np.abs(self.atomgroup.charges), compound=self.comp
        )
        testpos = center[self.inverse_ix]

        rbins = np.digitize(testpos, self._obs.bin_edges[1:-1])
        z = (np.arange(nbinsz)) * (self._obs.L / nbinsz)
        zbins = np.digitize(self.pos_cyl[self.atomgroup.ix, 2], z[1:])

        curQz = np.bincount(
            rbins + self.n_bins * zbins,
            weights=self.atomgroup.charges,
            minlength=self.n_bins * nbinsz,
        ).reshape(nbinsz, self.n_bins)

        curqz = np.cumsum(curQz, axis=0) / (self._obs.bin_area)[np.newaxis, :]
        self._obs.m_z = -curqz.mean(axis=0)
        # This is the systems dipole moment in z-direction and
        # not the radial integral of the dipole density.
        self._obs.M_z = np.dot(self._universe.atoms.charges, self.pos_cyl[:, 2])
        self._obs.mM_z = self._obs.m_z * self._obs.M_z

        # Save the total dipole moment in z dierection for correlation analysis.
        return self._obs.M_z

    def _conclude(self) -> None:
        super()._conclude()

        self._pref = 1 / scipy.constants.epsilon_0
        self._pref /= scipy.constants.Boltzmann * self.temperature
        # Convert from ~e^2/m to ~base units
        self._pref /= (
            scipy.constants.angstrom / (scipy.constants.elementary_charge) ** 2
        )

        if not self.single:
            # A factor of 2 pi L cancels out in the final expression because here M_z is
            # the total dipole moment in z-direction, not the radial integral of the
            # dipole density. M_z = 2 pi L \int_0^R dr r m(r)
            cov_z = self.means.mM_z - self.means.m_z * self.means.M_z
            cov_r = self.means.mM_r - self.means.m_r * self.means.M_r

            dcov_z = np.sqrt(
                self.sems.mM_z**2
                + self.sems.m_z**2 * self.means.M_z**2
                + self.means.m_z**2 * self.sems.M_z**2
            )
            dcov_r = np.sqrt(
                self.sems.mM_r**2
                + self.sems.m_r**2 * self.means.M_r**2
                + self.means.m_r**2 * self.sems.M_r**2
            )
        else:
            # <M> = 0 for a single line of water molecules.
            cov_z = self.means.mM_z
            cov_r = self.means.mM_r
            dcov_z = self.sems.mM_z
            dcov_r = self.sems.mM_r

        self.results.eps_z = self._pref * cov_z
        self.results.deps_z = self._pref * dcov_z

        self.results.eps_r = -(
            2 * np.pi * self._obs.L * self._pref * self.results.bin_pos * cov_r
        )
        self.results.deps_r = (
            2 * np.pi * self._obs.L * self._pref * self.results.bin_pos * dcov_r
        )

    @render_docs
    def save(self) -> None:
        """${SAVE_METHOD_PREFIX_DESCRIPTION}"""  # noqa: D415
        outdata_z = np.array(
            [self.results.bin_pos, self.results.eps_z, self.results.deps_z]
        ).T
        outdata_r = np.array(
            [self.results.bin_pos, self.results.eps_r, self.results.deps_r]
        ).T

        columns = ["positions [Å]"]

        columns += ["ε_z - 1", "Δε_z"]

        self.savetxt(
            "{}{}".format(self.output_prefix, "_z.dat"), outdata_z, columns=columns
        )

        columns = ["positions [Å]"]

        columns += ["ε^-1_r - 1", "Δε^-1_r"]

        self.savetxt(
            "{}{}".format(self.output_prefix, "_r.dat"), outdata_r, columns=columns
        )



#!/usr/bin/env python
#
# Copyright (c) 2026 Authors and contributors
# (see the AUTHORS.rst file for the full list of names)
#
# Released under the GNU Public Licence, v3 or any higher version
# SPDX-License-Identifier: GPL-3.0-or-later
r"""Module for computing 2D planar pair distribution functions."""

import logging

import MDAnalysis as mda
import numpy as np
from MDAnalysis.lib.distances import capped_distance

from ..core import PlanarBase
from ..lib.util import get_center, get_compound, render_docs

logger = logging.getLogger(__name__)


@render_docs
class PDFPlanar(PlanarBase):
    r"""Slab-wise planar 2D pair distribution functions.

    The pair distribution function :math:`g_\mathrm{2D}(r)` describes the
    spatial correlation between atoms in :math:`g_1` and atoms in
    :math:`g_2`, which lie in the same plane.
    It gives the average number density of :math:`g_2` atoms as a function of lateral
    distance :math:`r` from a centered :math:`g_1` atom.
    PDFPlanar can be used in systems that are inhomogeneous along one axis,
    and homogeneous in a plane.
    In fully homogeneous systems and in the limit of small 'dzheight'
    :math:`\Delta z`, it is the same as the well known three dimensional PDF.

    The planar PDF is defined by

    .. math::

        g_\mathrm{2D}(r) = \left \langle
        \frac{1}{N_{g1}} \cdot \sum_{i}^{N_{g1}} \sum_{j}^{N_{g2}}
        \frac{1}{2 \pi r} \delta(r - r_{ij}) \delta(z_{ij})
        \right \rangle .

    where the brackets :math:`\langle \cdot \rangle` denote the ensemble
    average. :math:`\delta(r- r_{ij})` counts the :math:`g_2` atoms at distance
    :math:`r` from atom :math:`i`.
    :math:`\delta(z_{ij})` ensures that only atoms, which lie
    in the same plane :math:`z_i = z_j`, are considered for the PDF.

    Discretized for computational purposes the equation reads as

    .. math::

        g_\mathrm{2D}(r) =
        \frac{1}{N_{g1}} \cdot \sum_{i}^{N_{g1}} \frac{\mathrm{count}\; g_2 \;
        \mathrm{in}\; \Delta V_i(r) }{\Delta V_i(r)} .

    where :math:`\Delta V_i(r)` is a ring around atom i, with inner
    radius :math:`r - \frac{\Delta r}{2}`, outer radius
    :math:`r + \frac{\Delta r}{2}` and height :math:`2 \Delta z`.

    As the density to normalise the PDF with is unknown, the output is in
    the dimension of number/volume in 1/Å^3.

    Functionally, PDFPlanar bins all pairwise :math:`g_1`-:math:`g_2` distances,
    where the z distance is smaller than 'dzheight' in a histogram.

    For a more detailed explanation refer to
    :ref:`Explanation: PDF<pdfs-explanation>` and
    :ref:`PDFPlanar Derivation<pdfplanar-derivation>`

    Parameters
    ----------
    ${PDF_PARAMETERS}
    pdf_bin_width : float
        Binwidth of bins in the histogram of the PDF (Å).
    dzheight : float
        dz height of a PDF slab :math:`\Delta z` (Å). :math:`\Delta z` is
        introduced to discretize the delta function :math:`\delta(z_{ij})`.
        It is the maximum :math:`z` distance between atoms which are
        considered to lie in the same plane.
        In the limit of :math:`\Delta z \to 0`, PDFPlanar reaches the
        continous limit. However, if :math:`\Delta z` is too small, there
        are no atoms in ``g2`` to sample.
        We recommend a choice of :math:`\Delta z` that is 1/10th of
        a bond length.
    dmin : float
        Minimum pairwise distance between ``g1`` and ``g2`` (Å).
    dmax : float
        Maximum pairwise distance between ``g1`` and ``g2`` (Å).
    ${PLANAR_CLASS_PARAMETERS}
    ${BIN_METHOD_PARAMETER}
    ${BASE_CLASS_PARAMETERS}
    ${OUTPUT_PARAMETER}

    Attributes
    ----------
    ${PLANAR_CLASS_ATTRIBUTES}
    results.bins: numpy.ndarray
        distances to which the PDF is calculated with shape (pdf_nbins) (Å)
    results.pdf: np.ndrray
        PDF with shape (pdf_nbins, n_bins) (1/Å^3)

    """

    def __init__(
        self,
        g1: mda.AtomGroup,
        g2: mda.AtomGroup | None = None,
        pdf_bin_width: float = 0.3,
        dzheight: float = 0.1,
        dmin: float = 0.0,
        dmax: float | None = None,
        dim: int = 2,
        zmin: float | None = None,
        zmax: float | None = None,
        bin_width: float = 1,
        bin_method: str = "com",
        refgroup: mda.AtomGroup | None = None,
        unwrap: bool = False,
        pack: bool = True,
        jitter: float = 0.0,
        concfreq: int = 0,
        output: str = "pdf.dat",
    ) -> None:
        self._locals = locals()
        self.comp_1 = get_compound(g1)
        super().__init__(
            atomgroup=g1,
            refgroup=refgroup,
            unwrap=unwrap,
            pack=pack,
            concfreq=concfreq,
            jitter=jitter,
            dim=dim,
            zmin=zmin,
            zmax=zmax,
            bin_width=bin_width,
            wrap_compound=self.comp_1,
        )

        self.g1 = g1
        if g2 is None:
            self.g2 = g1
        else:
            self.g2 = g2
        self.dmin = dmin
        self.dmax = dmax
        self.pdf_bin_width = pdf_bin_width
        self.dzheight = dzheight
        self.output = output
        self.bin_method = bin_method.lower()

        self.comp_2 = get_compound(self.g2)

    def _prepare(self) -> None:
        super()._prepare()
        logger.info("Analysis of the planar pair distribution function.")

        half_of_box_size = min(self.box_center)
        if self.dmax is None:
            self.dmax = min(self.box_center)
            logger.info(
                "Setting maximum range of PDF to half the box size ({self.range[1]} Å)."
            )
        elif self.dmax > min(self.box_center):
            raise ValueError(
                "Range of PDF exceeds half of the box size. Set to smaller than "
                f"{half_of_box_size} Å."
            )

        try:
            if self.pdf_bin_width > 0:
                self.pdf_nbins = int(
                    np.ceil((self.dmax - self.dmin) / self.pdf_bin_width)
                )
            else:
                raise ValueError("PDF bin_width must be a positive number.")
        except TypeError as err:
            raise ValueError("PDF bin_width must be a number.") from err

        if self.bin_method not in ["cog", "com", "coc"]:
            raise ValueError(
                f"{self.bin_method} is an unknown binning method. Use `cog`, `com` or "
                "`coc`."
            )

        logger.info(f"Using {self.pdf_nbins} pdf bins.")

        # Empty histogram self.count to store the PDF.
        self.edges = np.histogram(
            [-1], bins=self.pdf_nbins, range=(self.dmin, self.dmax)
        )[1]
        self.results.bins = 0.5 * (self.edges[:-1] + self.edges[1:])

        # Set the max range to filter the search radius.
        self._maxrange = self.dmax

    def _single_frame(self) -> None:
        super()._single_frame()
        self._obs.n_g1 = np.zeros((self.n_bins, 1))
        self._obs.count = np.zeros((self.n_bins, self.pdf_nbins))

        bin_width = (self.zmax - self.zmin) / self.n_bins

        g1_bin_positions = get_center(
            atomgroup=self.g1, bin_method=self.bin_method, compound=self.comp_1
        )
        g2_bin_positions = get_center(
            atomgroup=self.g2, bin_method=self.bin_method, compound=self.comp_2
        )

        # Calculate planar pdf per bin by averaging over all atoms in one bin.
        for z_bin in range(0, self.n_bins):
            # Set zmin and zmax of the bin.
            z_min = self.zmin + bin_width * z_bin
            z_max = self.zmin + bin_width * (z_bin + 1)

            # Get all atoms in a bin.
            g1_in_zbin_positions = g1_bin_positions[
                np.logical_and(
                    g1_bin_positions[:, self.dim] >= z_min,
                    g1_bin_positions[:, self.dim] < z_max,
                )
            ]

            g2_in_zbin_positions = g2_bin_positions[
                np.logical_and(
                    g2_bin_positions[:, self.dim] >= z_min - self.dzheight,
                    g2_bin_positions[:, self.dim] < z_max + self.dzheight,
                )
            ]

            n_g1 = len(g1_in_zbin_positions)
            n_g2 = len(g2_in_zbin_positions)
            self._obs.n_g1[z_bin] = n_g1

            # Extract z coordinate.
            z_g1 = np.copy(g1_in_zbin_positions)
            z_g2 = np.copy(g2_in_zbin_positions)
            # Set other coordinates to 0.
            z_g1[:, self.odims] = 0
            z_g2[:, self.odims] = 0

            # Automatically filter only those pairs with delta z < dz.
            z_pairs, _ = capped_distance(
                z_g1, z_g2, self.dzheight, box=self._universe.dimensions
            )

            # Calculate pairwise distances between g1 and g2.
            pairs, xy_distances = capped_distance(
                g1_in_zbin_positions,
                g2_in_zbin_positions,
                self._maxrange,
                box=self._universe.dimensions,
            )

            # Map pairs (i, j) to a number i+N*j (so we can use np.isin).
            z_pairs_encode = z_pairs[:, 0] + n_g2 * z_pairs[:, 1]
            pairs_encode = pairs[:, 0] + n_g2 * pairs[:, 1]

            mask_in_dz = np.isin(pairs_encode, z_pairs_encode)
            mask_different_atoms = np.where(xy_distances > 0, True, False)

            relevant_xy_distances = xy_distances[mask_in_dz * mask_different_atoms]
            # Histogram the pairwise distances.
            self._obs.count[z_bin] = np.histogram(
                relevant_xy_distances, bins=self.pdf_nbins, range=(self.dmin, self.dmax)
            )[0]

    def _conclude(self) -> None:
        super()._conclude()

        # Normalise pdf using the volumes of a ring with height 2*dz.
        ring_volumes = (
            np.pi * (self.edges[1:] ** 2 - self.edges[:-1] ** 2) * 2 * self.dzheight
        )
        ring_volumes = np.expand_dims(ring_volumes, axis=0)
        self.results.bins = self.results.bins
        self.results.pdf = self.means.count / self.means.n_g1 / ring_volumes
        self.results.pdf = np.nan_to_num(self.results.pdf.T, nan=0)

    @render_docs
    def save(self) -> None:
        """${SAVE_METHOD_DESCRIPTION}"""  # noqa: D415
        columns = ["r [Å]"]
        for z in self.results.bin_pos:
            columns.append(f"pdf at {z:.2f} Å [Å^-3]")

        self.savetxt(
            self.output,
            np.hstack([self.results.bins[:, np.newaxis], self.results.pdf]),
            columns=columns,
        )



#!/usr/bin/env python
#
# Copyright (c) 2026 Authors and contributors
# (see the AUTHORS.rst file for the full list of names)
#
# Released under the GNU Public Licence, v3 or any higher version
# SPDX-License-Identifier: GPL-3.0-or-later
r"""Module for computing planar dipolar order parameters."""

import logging

import MDAnalysis as mda

from ..core import ProfilePlanarBase
from ..lib.util import render_docs, unit_vectors_planar
from ..lib.weights import diporder_weights

logger = logging.getLogger(__name__)


@render_docs
class DiporderPlanar(ProfilePlanarBase):
    r"""Cartesian dipolar order parameters.

    ${DIPORDER_DESCRIPTION}

    ${CORRELATION_INFO_PLANAR}

    Parameters
    ----------
    ${ATOMGROUP_PARAMETER}
    ${ORDER_PARAMETER_PARAMETER}
    ${PDIM_PLANAR_PARAMETER}
    ${PROFILE_PLANAR_CLASS_PARAMETERS}
    ${OUTPUT_PARAMETER}

    Attributes
    ----------
    ${PROFILE_PLANAR_CLASS_ATTRIBUTES}

    """

    def __init__(
        self,
        atomgroup: mda.AtomGroup,
        order_parameter: str = "P0",
        pdim: int = 2,
        dim: int = 2,
        zmin: float | None = None,
        zmax: float | None = None,
        bin_width: float = 1,
        bin_method: str = "com",
        grouping: str = "residues",
        sym: bool = False,
        refgroup: mda.AtomGroup | None = None,
        unwrap: bool = True,
        pack: bool = True,
        jitter: float = 0.0,
        concfreq: int = 0,
        output: str = "diporder_planar.dat",
    ) -> None:
        self._locals = locals()
        normalization = "volume" if order_parameter == "P0" else "number"

        def get_unit_vectors(atomgroup: mda.AtomGroup, grouping: str):
            return unit_vectors_planar(
                atomgroup=atomgroup, grouping=grouping, pdim=pdim
            )

        super().__init__(
            atomgroup=atomgroup,
            unwrap=unwrap,
            pack=pack,
            refgroup=refgroup,
            jitter=jitter,
            concfreq=concfreq,
            dim=dim,
            zmin=zmin,
            zmax=zmax,
            bin_width=bin_width,
            sym=sym,
            sym_odd=True,
            grouping=grouping,
            bin_method=bin_method,
            output=output,
            weighting_function=diporder_weights,
            weighting_function_kwargs={
                "order_parameter": order_parameter,
                "get_unit_vectors": get_unit_vectors,
            },
            normalization=normalization,
        )

    def _prepare(self):
        logger.info("Analysis of the cartesian dipolar order parameters.")
        super()._prepare()



#!/usr/bin/env python
#
# Copyright (c) 2026 Authors and contributors
# (see the AUTHORS.rst file for the full list of names)
#
# Released under the GNU Public Licence, v3 or any higher version
# SPDX-License-Identifier: GPL-3.0-or-later
"""Module for computing planar density profiles."""

import logging

import MDAnalysis as mda

from ..core import ProfilePlanarBase
from ..lib.util import render_docs
from ..lib.weights import density_weights

logger = logging.getLogger(__name__)


@render_docs
class DensityPlanar(ProfilePlanarBase):
    r"""Cartesian partial density profiles.

    ${DENSITY_PLANAR_DESCRIPTION}

    ${CORRELATION_INFO_PLANAR}

    Parameters
    ----------
    ${ATOMGROUP_PARAMETER}
    ${DENS_PARAMETER}
    ${PROFILE_PLANAR_CLASS_PARAMETERS}
    ${OUTPUT_PARAMETER}

    Attributes
    ----------
    ${PROFILE_PLANAR_CLASS_ATTRIBUTES}

    Notes
    -----
    Partial mass density profiles can be used to calculate the ideal component of the
    chemical potential. For details, take a look at the corresponding :ref:`How-to
    guide<sphx_glr_generated_examples_basics_chemical-potential.py>`.

    """

    def __init__(
        self,
        atomgroup: mda.AtomGroup,
        dens: str = "mass",
        dim: int = 2,
        zmin: float | None = None,
        zmax: float | None = None,
        bin_width: float = 1,
        bin_method: str = "com",
        grouping: str = "atoms",
        sym: bool = False,
        refgroup: mda.AtomGroup | None = None,
        unwrap: bool = True,
        pack: bool = True,
        jitter: float = 0.0,
        concfreq: int = 0,
        output: str = "density.dat",
    ) -> None:
        self._locals = locals()
        super().__init__(
            atomgroup=atomgroup,
            unwrap=unwrap,
            pack=pack,
            jitter=jitter,
            concfreq=concfreq,
            dim=dim,
            zmin=zmin,
            zmax=zmax,
            bin_width=bin_width,
            refgroup=refgroup,
            sym=sym,
            sym_odd=False,
            grouping=grouping,
            bin_method=bin_method,
            output=output,
            weighting_function=density_weights,
            weighting_function_kwargs={"dens": dens},
            normalization="volume",
        )

    def _prepare(self):
        logger.info(f"Analysis of the {self._locals['dens']} density profile.")
        super()._prepare()



#!/usr/bin/env python
#
# Copyright (c) 2026 Authors and contributors
# (see the AUTHORS.rst file for the full list of names)
#
# Released under the GNU Public Licence, v3 or any higher version
# SPDX-License-Identifier: GPL-3.0-or-later
"""Base class for planar analysis."""

import logging
from collections.abc import Callable

import MDAnalysis as mda
import numpy as np

from ..lib.math import symmetrize
from ..lib.util import render_docs
from .base import AnalysisBase, ProfileBase

logger = logging.getLogger(__name__)


@render_docs
class PlanarBase(AnalysisBase):
    r"""Analysis class providing options and attributes for a planar system.

    Parameters
    ----------
    ${ATOMGROUP_PARAMETER}
    ${PLANAR_CLASS_PARAMETERS}
    ${WRAP_COMPOUND_PARAMETER}

    Attributes
    ----------
    ${PLANAR_CLASS_ATTRIBUTES}
    zmin : float
         Minimal coordinate for evaluation (Å) with in the lab frame, where 0
         corresponds to the origin of the cell.
    zmax : float
         Maximal coordinate for evaluation (Å) with in the lab frame, where 0
         corresponds to the origin of the cell.
    _obs.L : float
        Length (in Å) along the chosen dimension in the current frame.
    _obs.bin_pos : numpy.ndarray, (n_bins)
        Central bin positions (in Å) of each bin (in Å) in the current frame.
    _obs.bin_width : float
         Bin width (in Å) in the current frame
    _obs.bin_edges : numpy.ndarray, (n_bins + 1)
        Edges of the bins (in Å) in the current frame.
    _obs.bin_area : numpy.ndarray, (n_bins)
        Area of the rectangle of each bin in the current frame. Calculated via
        :math:`L_x \cdot L_y / N_\mathrm{bins}` where :math:`L_x` and :math:`L_y` are
        the box lengths perpendicular to the dimension of evaluations given by ``dim``.
        :math:`N_\mathrm{bins}` is the number of bins.
    _obs.bin_volume : numpy.ndarray, (n_bins)
        Volume of an cuboid of each bin (in Å^3) in the current frame.

    """

    def __init__(
        self,
        atomgroup: mda.AtomGroup,
        unwrap: bool,
        pack: bool,
        refgroup: mda.AtomGroup | None,
        jitter: float,
        concfreq: int,
        dim: int,
        zmin: None | float,
        zmax: None | float,
        bin_width: float,
        wrap_compound: str,
    ):
        super().__init__(
            atomgroup=atomgroup,
            unwrap=unwrap,
            pack=pack,
            refgroup=refgroup,
            jitter=jitter,
            concfreq=concfreq,
            wrap_compound=wrap_compound,
        )

        if dim not in [0, 1, 2]:
            raise ValueError("Dimension can only be x=0, y=1 or z=2.")
        self.dim = dim

        # These values are requested by the user, but the actual ones are calculated
        # during runtime in the lab frame
        self._zmax = zmax
        self._zmin = zmin
        self._bin_width = bin_width

    @property
    def odims(self) -> np.ndarray:
        """Other dimensions perpendicular to dim i.e. (0,2) if dim = 1."""
        return np.roll(np.arange(3), -self.dim)[1:]

    def _compute_lab_frame_planar(self):
        """Compute lab limits `zmin` and `zmax`."""
        current_box_length = self.box_lengths[self.dim]
        if (
            np.abs(current_box_length - self.initial_box_length)
            / self.initial_box_length
            > 0.05
            and not self._warned_box_length
        ):
            logger.warning(
                f"Box length along dimension {self.dim} has changed more than 5 % "
                "since the start of the analysis. This may lead to unexpected "
                "behavior due to a fixed number of bins."
            )
            self._warned_box_length = True

        if self._zmin is None:
            self.zmin = 0
        elif abs(self._zmin) > current_box_length / 2 and not self._warned_zmin:
            logger.warning(
                f"User-defined zmin ({self._zmin:.2f} Å) exceeds half the current "
                f"box length ({current_box_length / 2:.2f} Å) along dimension "
                f"{self.dim}. Consider letting MAICoS calculate bounds automatically "
                "by setting zmin to `None` or manually determine safe bounds from the "
                "smallest box vector across frames."
            )
            self._warned_zmin = True
            self.zmin = self.box_center[self.dim] + self._zmin
        else:
            self.zmin = self.box_center[self.dim] + self._zmin

        if self._zmax is None:
            self.zmax = self.box_lengths[self.dim]
        elif abs(self._zmax) > current_box_length / 2 and not self._warned_zmax:
            logger.warning(
                f"User-defined zmax ({self._zmax:.2f} Å) exceeds half the current box "
                f"length ({current_box_length / 2:.2f} Å) along dimension {self.dim}. "
                "Consider letting MAICoS calculate bounds automatically (zmax=None) "
                "or manually determine safe bounds from the smallest box vector "
                "across frames."
            )
            self._warned_zmax = True
            self.zmax = self.box_center[self.dim] + self._zmax
        else:
            self.zmax = self.box_center[self.dim] + self._zmax
        # enforce calculations in double precision
        self.zmin = np.float64(self.zmin)
        self.zmax = np.float64(self.zmax)

    def _prepare(self):
        """Prepare the planar analysis."""
        self._warned_zmin = False
        self._warned_zmax = False
        self._warned_box_length = False
        self.initial_box_length = self.box_lengths[self.dim]
        self._compute_lab_frame_planar()
        # TODO(@hejamu): There are much more wrong combinations of zmin and zmax...
        if (
            self._zmax is not None
            and self._zmin is not None
            and self._zmax <= self._zmin
        ):
            raise ValueError("`zmax` can not be smaller or equal than `zmin`!")

        L = self.zmax - self.zmin
        self.n_bins = int(np.ceil(L / self._bin_width))

    def _single_frame(self):
        """Single frame for the planar analysis."""
        self._compute_lab_frame_planar()
        self._obs.L = self.zmax - self.zmin
        self._obs.box_center = self.box_center
        self._obs.bin_edges = np.linspace(self.zmin, self.zmax, self.n_bins + 1)

        self._obs.bin_width = self._obs.L / self.n_bins
        self._obs.bin_pos = self._obs.bin_edges[1:] - self._obs.bin_width / 2
        # We define `bin_area` and `bin_volume` as array of length `n_bins` even though
        # each element has the same value. With this the array shape is consistent with
        # the cylindrical and spherical classes, where `bin_area` and `bin_volume` is
        # different in each bin.
        self._obs.bin_area = np.ones(self.n_bins) * np.prod(
            self.box_lengths[self.odims]
        )
        self._obs.bin_volume = self._obs.bin_area * self._obs.bin_width

    def _conclude(self):
        """Results calculations for the planar analysis."""
        # Convert coordinates back from lab frame to refgroup frame.
        self.results.bin_pos = self.means.bin_pos - self.means.box_center[self.dim]


@render_docs
class ProfilePlanarBase(PlanarBase, ProfileBase):  # type: ignore
    """Base class for computing profiles in a cartesian geometry.

    ${CORRELATION_INFO_RADIAL}

    Parameters
    ----------
    ${PROFILE_PLANAR_CLASS_PARAMETERS}
    ${PROFILE_CLASS_PARAMETERS_PRIVATE}

    Attributes
    ----------
    ${PROFILE_PLANAR_CLASS_ATTRIBUTES}

    """

    def __init__(
        self,
        atomgroup: mda.AtomGroup,
        unwrap: bool,
        pack: bool,
        refgroup: mda.AtomGroup | None,
        jitter: float,
        concfreq: int,
        dim: int,
        zmin: None | float,
        zmax: None | float,
        bin_width: float,
        sym: bool,
        sym_odd: bool,
        grouping: str,
        bin_method: str,
        output: str,
        weighting_function: Callable,
        weighting_function_kwargs: None | dict,
        normalization: str,
    ):
        PlanarBase.__init__(
            self,
            atomgroup=atomgroup,
            unwrap=unwrap,
            pack=pack,
            refgroup=refgroup,
            jitter=jitter,
            concfreq=concfreq,
            dim=dim,
            zmin=zmin,
            zmax=zmax,
            bin_width=bin_width,
            wrap_compound=grouping,  # same as grouping to avoid broken compounds
        )
        # `AnalysisBase` performs conversions on `atomgroup`.
        # Take converted `atomgroup` and not the user provided ones.
        ProfileBase.__init__(
            self,
            atomgroup=self.atomgroup,
            grouping=grouping,
            bin_method=bin_method,
            output=output,
            weighting_function=weighting_function,
            weighting_function_kwargs=weighting_function_kwargs,
            normalization=normalization,
        )

        self.sym = sym

        if self.sym and self.refgroup is None:
            raise ValueError("For symmetrization the `refgroup` argument is required.")

        self.sym_odd = sym_odd

    def _prepare(self):
        PlanarBase._prepare(self)
        ProfileBase._prepare(self)

        logger.info(f"""Profile along {"xyz"[self.dim]}-axis normal to the plane.""")

    def _compute_histogram(
        self, positions: np.ndarray, weights: np.ndarray | None = None
    ) -> np.ndarray:
        positions = positions[:, self.dim]
        hist, bin_edges = np.histogram(
            positions, bins=self.n_bins, range=(self.zmin, self.zmax), weights=weights
        )
        # TODO(@hejamu): Is this the best way to do this?
        # Also, can we somehow abstract this away?
        bin_indices = np.digitize(positions, bin_edges) - 1
        bin_indices[bin_indices == self.n_bins] = -1

        return hist, bin_indices

    def _single_frame(self) -> float:
        PlanarBase._single_frame(self)
        ProfileBase._single_frame(self)

        # Take the center bin for correlation analysis.
        return self._obs.profile[self.n_bins // 2]

    def _conclude(self):
        PlanarBase._conclude(self)
        ProfileBase._conclude(self)
        if self.sym:
            symmetrize(self.sums.profile, inplace=True, is_odd=self.sym_odd)
            symmetrize(self.means.profile, inplace=True, is_odd=self.sym_odd)

            symmetrize(self.sems.profile, inplace=True, is_odd=False)

            if self.normalization == "number":
                symmetrize(self.sums.bincount, inplace=True, is_odd=self.sym_odd)



#!/usr/bin/env python
#
# Copyright (c) 2026 Authors and contributors
# (see the AUTHORS.rst file for the full list of names)
#
# Released under the GNU Public Licence, v3 or any higher version
# SPDX-License-Identifier: GPL-3.0-or-later
"""Base class for spherical analysis."""

import logging
from collections.abc import Callable

import MDAnalysis as mda
import numpy as np

from ..lib.math import transform_sphere
from ..lib.util import render_docs
from .base import ProfileBase
from .planar import AnalysisBase

logger = logging.getLogger(__name__)


@render_docs
class SphereBase(AnalysisBase):
    r"""Analysis class providing options and attributes for a spherical system.

    Provide the results attribute ``r``.

    Parameters
    ----------
    ${ATOMGROUP_PARAMETER}
    ${SPHERE_CLASS_PARAMETERS}
    ${WRAP_COMPOUND_PARAMETER}

    Attributes
    ----------
    ${SPHERE_CLASS_ATTRIBUTES}
    pos_sph : numpy.ndarray
        positions in spherical coordinats ``(r, phi, theta)``
    _obs.R : float
        Average length (in Å) along the radial dimension in the current frame.
    _obs.bin_pos : numpy.ndarray, (n_bins)
        Central bin position of each bin (in Å) in the current frame.
    _obs.bin_width : float
         Bin width (in Å) in the current frame
    _obs.bin_edges : numpy.ndarray, (n_bins + 1)
        Edges of the bins (in Å) in the current frame.
    _obs.bin_area : numpy.ndarray, (n_bins)
        Surface area (in Å^2) of the sphere of each bin with radius ``bin_pos`` in the
        current frame. Calculated via :math:`4 \pi r_i^2` where :math:`i` is the index
        of the bin.
    results.bin_volume : numpy.ndarray, (n_bins)
        volume of a spherical shell of each bins (in Å^3) of the current frame.
        Calculated via :math:`4\pi/3 \left(r_{i+1}^3 - r_i^3 \right)` where :math:`i` is
        the index of the bin.

    """

    def __init__(
        self,
        atomgroup: mda.AtomGroup,
        unwrap: bool,
        pack: bool,
        refgroup: mda.AtomGroup | None,
        jitter: float,
        concfreq: int,
        rmin: float,
        rmax: None | float,
        bin_width: float,
        wrap_compound: str,
    ):
        super().__init__(
            atomgroup=atomgroup,
            unwrap=unwrap,
            pack=pack,
            refgroup=refgroup,
            jitter=jitter,
            concfreq=concfreq,
            wrap_compound=wrap_compound,
        )

        self.rmin = rmin
        self._rmax = rmax
        self._bin_width = bin_width

    def _compute_lab_frame_sphere(self):
        """Compute lab limit `rmax`."""
        box_half = self.box_lengths.min() / 2
        if self._rmax is None:
            self.rmax = box_half
        elif self._rmax <= box_half:
            self.rmax = self._rmax
        else:
            logger.warning(
                f"`rmax` is bigger than half the smallest box vector ({box_half:.2f}) "
                "in the radial direction. This will lead to artifacts at the edges."
            )
            self.rmax = self._rmax
        # enforce calculations in double precision
        self.rmin = np.float64(self.rmin)
        self.rmax = np.float64(self.rmax)
        # Transform into spherical coordinates
        self.pos_sph = transform_sphere(
            self._universe.atoms.positions, origin=self.box_center
        )

    def _prepare(self):
        """Prepare the spherical analysis."""
        self._compute_lab_frame_sphere()

        if self.rmin < 0:
            raise ValueError("Only values for `rmin` larger or equal 0 are allowed.")

        if self._rmax is not None and self._rmax <= self.rmin:
            raise ValueError("`rmax` can not be smaller than or equal to `rmin`!")

        R = self.rmax - self.rmin
        self.n_bins = int(np.ceil(R / self._bin_width))

    def _single_frame(self):
        """Single frame for the sphercial analysis."""
        self._compute_lab_frame_sphere()
        self._obs.R = self.rmax - self.rmin

        self._obs.bin_edges = np.linspace(
            self.rmin, self.rmax, self.n_bins + 1, endpoint=True
        )

        self._obs.bin_width = self._obs.R / self.n_bins
        self._obs.bin_pos = self._obs.bin_edges[1:] - self._obs.bin_width / 2
        self._obs.bin_area = 4 * np.pi * self._obs.bin_pos**2
        self._obs.bin_volume = 4 * np.pi * np.diff(self._obs.bin_edges**3) / 3

    def _conclude(self):
        """Results calculations for the sphercial analysis."""
        super()._conclude()
        self.results.bin_pos = self.means.bin_pos


@render_docs
class ProfileSphereBase(SphereBase, ProfileBase):  # type: ignore
    """Base class for computing radial profiles in a spherical geometry.

    ${CORRELATION_INFO_RADIAL}

    Parameters
    ----------
    ${PROFILE_SPHERE_CLASS_PARAMETERS}
    ${PROFILE_CLASS_PARAMETERS_PRIVATE}

    Attributes
    ----------
    ${PROFILE_CYLINDER_CLASS_ATTRIBUTES}

    """

    def __init__(
        self,
        atomgroup: mda.AtomGroup,
        unwrap: bool,
        pack: bool,
        refgroup: mda.AtomGroup | None,
        jitter: float,
        concfreq: int,
        rmin: float,
        rmax: None | float,
        bin_width: float,
        grouping: str,
        bin_method: str,
        output: str,
        weighting_function: Callable,
        weighting_function_kwargs: dict | None,
        normalization: str,
    ):
        SphereBase.__init__(
            self,
            atomgroup=atomgroup,
            unwrap=unwrap,
            pack=pack,
            refgroup=refgroup,
            jitter=jitter,
            concfreq=concfreq,
            rmin=rmin,
            rmax=rmax,
            bin_width=bin_width,
            wrap_compound=grouping,  # same as grouping to avoid broken compounds
        )
        # `AnalysisBase` performs conversions on `atomgroup`.
        # Take converted `atomgroup` and not the user provided ones.
        ProfileBase.__init__(
            self,
            atomgroup=self.atomgroup,
            grouping=grouping,
            bin_method=bin_method,
            output=output,
            weighting_function=weighting_function,
            weighting_function_kwargs=weighting_function_kwargs,
            normalization=normalization,
        )

    def _prepare(self):
        SphereBase._prepare(self)
        ProfileBase._prepare(self)

        logger.info(
            """Profile along the radial coordinate in a spherical """
            """coordinate system."""
        )

    def _compute_histogram(
        self, positions: np.ndarray, weights: np.ndarray | None = None
    ) -> np.ndarray:
        positions = transform_sphere(positions, origin=self.box_center)[:, 0]
        hist, bin_edges = np.histogram(
            positions, bins=self.n_bins, range=(self.rmin, self.rmax), weights=weights
        )

        # TODO(@hejamu): Is this the best way to do this?
        # Also, can we somehow abstract this away?
        bin_indices = np.digitize(positions, bin_edges) - 1
        bin_indices[bin_indices == self.n_bins] = -1

        return hist, bin_indices

    def _single_frame(self) -> float:
        SphereBase._single_frame(self)
        ProfileBase._single_frame(self)

        # Take the center bin of the zeroth group for correlation analysis.
        return self._obs.profile[0]

    def _conclude(self):
        SphereBase._conclude(self)
        ProfileBase._conclude(self)



#!/usr/bin/env python
#
# Copyright (c) 2026 Authors and contributors
# (see the AUTHORS.rst file for the full list of names)
#
# Released under the GNU Public Licence, v3 or any higher version
# SPDX-License-Identifier: GPL-3.0-or-later
"""Module for calculating spherical dielectric profiles."""

import logging

import MDAnalysis as mda
import numpy as np
import scipy.constants

from ..core import SphereBase
from ..lib.util import charge_neutral, citation_reminder, get_compound, render_docs

logger = logging.getLogger(__name__)


@render_docs
@charge_neutral(filter="error")
class DielectricSphere(SphereBase):
    r"""Spherical dielectric profiles.

    Computes the inverse radial :math:`\varepsilon_r^{-1}(r)` component of the
    spherical dielectric tensor :math:`\varepsilon`. The center of the sphere is either
    located at the center of the simulation box (default) or at the center of mass of
    the ``refgroup``, if provided.

    For usage, please refer to
    :ref:`sphx_glr_generated_examples_dielectrics_dielectric-profiles.py` and for
    details on the theory see :ref:`dielectric-explanations`.

    For correlation analysis, the radial (:math:`r`) component is used.
    ${CORRELATION_INFO}

    Also, please read and cite :footcite:p:`schaafDielectricResponseWater2015`.

    Parameters
    ----------
    ${ATOMGROUP_PARAMETER}
    ${TEMPERATURE_PARAMETER}
    ${SPHERE_CLASS_PARAMETERS}
    ${BASE_CLASS_PARAMETERS}
    ${OUTPUT_PREFIX_PARAMETER}

    Attributes
    ----------
    ${RADIAL_CLASS_ATTRIBUTES}
    results.eps_rad : numpy.ndarray
        Reduced inverse radial dielectric profile (:math:`\varepsilon^{-1}_r(r) - 1)`
    results.deps_rad : numpy.ndarray
        Uncertainty of inverse radial dielectric profile


    """

    def __init__(
        self,
        atomgroup: mda.AtomGroup,
        temperature: float = 300,
        rmin: float = 0,
        rmax: float | None = None,
        bin_width: float = 0.1,
        refgroup: mda.AtomGroup | None = None,
        unwrap: bool = True,
        pack: bool = True,
        jitter: float = 0.0,
        concfreq: int = 0,
        output_prefix: str = "eps_sph",
    ) -> None:
        self._locals = locals()
        self.comp = get_compound(atomgroup)
        ix = atomgroup._get_compound_indices(self.comp)
        _, self.inverse_ix = np.unique(ix, return_inverse=True)
        if rmin != 0 or rmax is not None:
            logger.warning(
                "Setting `rmin` and `rmax` might cut off molecules. This will lead to "
                "severe artifacts in the dielectric profiles."
            )

        super().__init__(
            atomgroup,
            concfreq=concfreq,
            jitter=jitter,
            refgroup=refgroup,
            rmin=rmin,
            rmax=rmax,
            bin_width=bin_width,
            unwrap=unwrap,
            pack=pack,
            wrap_compound=self.comp,
        )
        self.output_prefix = output_prefix
        self.bin_width = bin_width
        self.temperature = temperature

    def _prepare(self) -> None:
        logger.info(
            "Analysis of the inverse radial component "
            "of the spherical dielectric tensor."
        )
        # Print the Christian Schaaf citation
        logger.info(citation_reminder("10.1103/PhysRevE.92.032718"))

        super()._prepare()

    def _single_frame(self) -> float:
        super()._single_frame()

        # Precalculate the bins each atom belongs to.
        rbins = np.digitize(self.pos_sph[:, 0], self._obs.bin_edges[1:-1])

        # Calculate the charge per bin for the selected atomgroup.
        curQ_rad = np.bincount(
            rbins[self.atomgroup.ix],
            weights=self.atomgroup.charges,
            minlength=self.n_bins,
        )

        # In literature, the charge density is integrated along the radial direction to
        # get the dipole moment density. We can rewrite the integral by identifying:
        # q(a) = 4 * pi * int_0^a * r^2 * ρ(r) dr,
        # where q(a) is the charge enclosed within a sphere of radius a. This allows us
        # to avoid numerical errors.
        self._obs.m_r = -np.cumsum(curQ_rad) / 4 / np.pi / self._obs.bin_pos**2

        curQ_rad_tot = np.bincount(
            rbins, weights=self._universe.atoms.charges, minlength=self.n_bins
        )

        # Same as above, but for the total charge density.
        self._obs.m_r_tot = -np.cumsum(curQ_rad_tot) / 4 / np.pi / self._obs.bin_pos**2

        # This is not really the systems dipole moment, but it keeps the nomenclature
        # consistent with the DielectricPlanar module.
        self._obs.M_r = np.sum(self._obs.m_r_tot * self._obs.bin_width)
        self._obs.mM_r = self._obs.m_r * self._obs.M_r

        return self._obs.M_r

    def _conclude(self) -> None:
        super()._conclude()

        self._pref = 1 / scipy.constants.epsilon_0
        self._pref /= scipy.constants.Boltzmann * self.temperature
        # Convert from ~e^2/m to ~base units
        self._pref /= (
            scipy.constants.angstrom / (scipy.constants.elementary_charge) ** 2
        )

        cov_rad = self.means.mM_r - self.means.m_r * self.means.M_r

        dcov_rad = np.sqrt(
            self.sems.mM_r**2
            + self.sems.m_r**2 * self.means.M_r**2
            + self.means.m_r**2 * self.sems.M_r**2
        )

        self.results.eps_rad = 1 - (
            4 * np.pi * self.results.bin_pos**2 * self._pref * cov_rad
        )
        self.results.deps_rad = (
            4 * np.pi * self.results.bin_pos**2 * self._pref * dcov_rad
        )

    @render_docs
    def save(self) -> None:
        """${SAVE_METHOD_PREFIX_DESCRIPTION}"""  # noqa: D415
        outdata_rad = np.array(
            [self.results.bin_pos, self.results.eps_rad, self.results.deps_rad]
        ).T

        columns = ["positions [Å]", "eps_rad - 1", "eps_rad error"]

        self.savetxt(
            "{}{}".format(self.output_prefix, "_rad.dat"), outdata_rad, columns=columns
        )



# -*-
#
# Copyright (c) 2026 Authors and contributors (see the AUTHORS.rst file for the full
# list of names)
#
# Released under the GNU Public Licence, v3 or any higher version
# SPDX-License-Identifier: GPL-3.0-or-later
"""Helper functions for mathematical and physical operations."""

import MDAnalysis as mda
import numpy as np

# Max spacing variation in series that is allowed
dt_dk_tolerance = 1e-8  # (~1e-10 suggested)
dr_tolerance = 1e-6


def correlation(
    a: np.ndarray, b: np.ndarray | None = None, subtract_mean: bool = False
) -> np.ndarray:
    """Calculate correlation or autocorrelation.

    Uses fast fourier transforms to give the correlation function of two arrays, or, if
    only one array is given, the autocorrelation. Setting ``subtract_mean=True`` causes
    the mean to be subtracted from the input data.

    Parameters
    ----------
    a : numpy.ndarray
        The first input array to calculate the correlation
    b : numpy.ndarray
        The second input array. If :obj:`None`, autocorrelation of ``a`` is calculated.
    subtract_mean : bool
        If :obj:`True`, subtract the mean from the input data.

    Returns
    -------
    numpy.ndarray
        The correlation or autocorrelation function.

    """
    meana = int(subtract_mean) * np.mean(
        a
    )  # essentially an if statement for subtracting mean
    a2 = np.append(
        a - meana, np.zeros(2 ** int(np.ceil(np.log(len(a)) / np.log(2))) - len(a))
    )  # round up to a power of 2
    data_a = np.append(a2, np.zeros(len(a2)))  # pad with an equal number of zeros
    fra = np.fft.fft(data_a)  # FT the data
    if b is None:
        sf = (
            np.conj(fra) * fra
        )  # take the conj and multiply pointwise if autocorrelation
    else:
        meanb = int(subtract_mean) * np.mean(b)
        b2 = np.append(
            b - meanb,
            np.zeros(2 ** int(np.ceil(np.log(len(b)) / np.log(2))) - len(b)),
        )
        data_b = np.append(b2, np.zeros(len(b2)))
        frb = np.fft.fft(data_b)
        sf = np.conj(fra) * frb
    return np.real(np.fft.ifft(sf)[: len(a)]) / np.array(
        range(len(a), 0, -1)
    )  # inverse FFT and normalization


def scalar_prod_corr(
    a: np.ndarray, b: np.ndarray | None = None, subtract_mean: bool = False
) -> np.ndarray:
    """Give the corr. function of the scalar product of two vector timeseries.

    Arguments should be given in the form a[t, i], where t is the time variable along
    which the correlation is calculated, and i indexes the vector components.

    Parameters
    ----------
    a : numpy.ndarray
        The first vector timeseries of shape (t, i).
    b : numpy.ndarray
        The second vector timeseries of shape (t, i). If :obj:`None`, correlation with
        itself is calculated.
    subtract_mean : bool
        If :obj:`True`, subtract the mean from the timeseries before calculating the
        correlation.

    Returns
    -------
    numpy.ndarray
        The correlation function of the scalar product of the vector timeseries.

    """
    corr = np.zeros(len(a[:, 0]))

    if b is None:
        for i in range(0, len(a[0, :])):
            corr[:] += correlation(a[:, i], None, subtract_mean)

    else:
        for i in range(0, len(a[0, :])):
            corr[:] += correlation(a[:, i], b[:, i], subtract_mean)

    return corr


def correlation_time(
    timeseries: np.ndarray,
    method: str = "sokal",
    mintime: int = 3,
    sokal_factor: float = 8,
) -> float:
    r"""Compute the integrated correlation time of a time series.

    The integrated correlation time (in units of the sampling interval) is given by

    .. math::
        \tau = \sum\limits_{t=1}^{N_\mathrm{cut}} C(t) \left(1 - \frac{t}{N}\right)

    where :math:`N_\mathrm{cut} < N` is a subset of the time series of length :math:`N`
    and :math:`C(t)` is the discrete-time autocorrelation function. To obtain the upper
    limit of the sum :math:`N_\mathrm{cut}` two different methods are provided:

    1. For "chodera" :footcite:p:`choderaWeightedHistogramAnalysis2007`
       :math:`N_\mathrm{cut}` is given by the time when :math:`C(t)`
       crosses zero the first time.

    2. For "sokal" :footcite:p:`sokalLecture` :math:`N_\mathrm{cut}` is determined
       iteratively by stepwise increasing until

       .. math::
            N_\mathrm{cut} \geq c \cdot \tau

       where :math:`c` is the constant ``sokal_factor``. If the condition is never
       fulfilled, ``-1`` is returned, indicating that the time series does not provide
       sufficient statistics to estimate a
       correlation time.

    While both methods give the same correlation time for a smooth time series that
    decays to 0, "sokal" will results in a more reasonable result for actual time series
    that are noisy and cross zero several times.

    Parameters
    ----------
    timeseries : numpy.ndarray
        The time series used to calculate the correlation time from.
    method : {``"sokal"``, ``"chodera"``}
        Method to choose summation cutoff :math:`N_\mathrm{cut}`.
    mintime: int
        Minimum possible value for :math:`N_\mathrm{cut}`.
    sokal_factor : float
        Cut-off factor :math:`c` for the Sokal method.

    Returns
    -------
    tau : float
        Integrated correlation time :math:`\tau`. If ``-1`` (only for
        ``method="sokal"``) the provided time series does not provide sufficient
        statistics to estimate a correlation time.

    Raises
    ------
    ValueError
        If mintime is larger than the length of the timeseries.
    ValueError
        If method is not one of "sokal" or "chodera".

    References
    ----------
    .. footbibliography::

    """
    if mintime > len(timeseries):
        raise ValueError(
            f"mintime ({mintime}) has to be smaller then the length of `timeseries` "
            f"({len(timeseries)})."
        )

    corr = correlation(timeseries, subtract_mean=True)

    if method == "sokal":
        for cutoff in range(mintime, len(timeseries)):
            tau = np.sum(
                (1 - np.arange(1, cutoff) / len(timeseries)) * corr[1:cutoff] / corr[0]
            )
            if cutoff >= sokal_factor * tau:
                break

            if cutoff > len(timeseries) / 3:
                return -1

    elif method == "chodera":
        cutoff = np.max([mintime, np.min(np.argwhere(corr < 0))])
        tau = np.sum(
            (1 - np.arange(1, cutoff) / len(timeseries)) * corr[1:cutoff] / corr[0]
        )
    else:
        raise ValueError(
            f"Unknown method: {method}. Chose either 'sokal' or 'chodera'."
        )
    return tau


def new_mean(old_mean: float, data: float, length: int) -> float:
    r"""Compute the arithmetic mean of a series iteratively.

    Compute the arithmetic mean of n samples based on an existing mean of n-1 and the
    n-th value.

    Given the mean of a data series

    .. math::

        \bar x_N = \frac{1}{N} \sum_{n=1}^N x_n

    we seperate the last value

    .. math::

        \bar x_N = \frac{1}{N} \sum_{n=1}^{N-1} x_n + \frac{x_N}{N}

    and multiply 1 = (N - 1)/(N - 1)

    .. math::

        \bar x_N = \frac{N-1}{N} \frac{1}{N-1} \\ \sum_{n=1}^{N-1} x_n + \frac{x_N}{N}

    The first term can be identified as the mean of the first N - 1 values and we arrive
    at

    .. math::

        \bar x_N = \frac{N-1}{N} \bar x_{N-1} + \frac{x_N}{N}


    Parameters
    ----------
    old_mean : float
        arithmetic mean of the first n - 1 samples.
    data : float
        n-th value of the series.
    length : int
        Length of the updated series, here called n.

    Returns
    -------
    new_mean : float
        Updated mean of the series of n values.

    Examples
    --------
    The mean of a data set can easily be calculated from the data points. However this
    requires one to keep all data points on hand until the end of the calculation.

    >>> print(np.mean([1, 3, 5, 7]))
    4.0

    Alternatively, one can update an existing mean, this requires only knowledge of the
    total number of samples.

    >>> print(new_mean(np.mean([1, 3, 5]), data=7, length=4))
    4.0

    """
    return ((length - 1) * old_mean + data) / length


def new_variance(
    old_variance: float | np.ndarray,
    old_mean: float | np.ndarray,
    new_mean: float | np.ndarray,
    data: float | np.ndarray,
    length: int,
) -> float | np.ndarray:
    r"""Calculate the variance of a timeseries iteratively.

    The variance of a timeseries :math:`x_n` can be calculated iteratively by using the
    following formula:

    .. math::

        S_n = S_n-1 + (n-1) * (x_n - \bar{x}_n-1)^2 / (n-1)

    Here, :math:`\bar{x}_n` is the mean of the timeseries up to the :math:`n`-th value.

    Floating point imprecision can lead to slight negative variances leading non defined
    standard deviations. Therefore a negetaive variance is set to 0.

    Parameters
    ----------
    old_variance : float, numpy.ndarray
        The variance of the first n-1 samples.
    old_mean : float
        The mean of the first n-1 samples.
    new_mean : float, numpy.ndarray
        The mean of the full n samples.
    data : float, numpy.ndarray
        The n-th value of the series.
    length : int
        Length of the updated series, here called n.

    Returns
    -------
    new_variance : float
        Updated variance of the series of n values.

    Examples
    --------
    The data set ``[1, 5, 5, 1]`` has a variance of ``4.0``

    >>> print(np.var([1, 5, 5, 1]))
    4.0

    Knowing the total number of data points, this operation can be performed
    iteratively.

    >>> print(
    ...     new_variance(
    ...         old_variance=np.var([1, 5, 5]),
    ...         old_mean=np.mean([1, 5, 5]),
    ...         new_mean=np.mean([1, 5, 5, 1]),
    ...         data=1,
    ...         length=4,
    ...     )
    ... )
    4.0

    """
    S_old = old_variance * (length - 1)
    S_new = S_old + (data - old_mean) * (data - new_mean)

    if type(S_new) is np.ndarray:
        S_new[S_new < 0] = 0
    else:
        if S_new < 0:
            S_new = 0

    return S_new / length


def center_cluster(ag: mda.AtomGroup, weights: np.ndarray) -> np.ndarray:
    """Calculate the center of the atomgroup with respect to some weights.

    Parameters
    ----------
    ag : MDAnalysis.core.groups.AtomGroup
        Group of atoms to calculate the center for.

    weights : numpy.ndarray
        Weights in the shape of ag.

    Returns
    -------
    com : numpy.ndarray
        The center with respect to the weights.


    Without proper treatment of periodic boundrary conditions (PBC) most algorithms will
    result in wrong center calculations. As shown below without treating PBC the center
    of mass is located in the center of the box ::

       +-----------+
       |           |
       | 1   x   2 |
       |           |
       +-----------+

    However, the distance accross the box boundary is shorter and therefore the center
    with PBC should be located somwhere else. The correct way to calculate the center is
    described in :footcite:t:`bai_calculating_2008` where coordinates of the particles
    are projected on a circle and weighted by their mass in this two dimensional space.
    The center of mass is obtained by transforming this point back to the corresponding
    point in the real system. This is done seperately for each dimension.

    Reasons for doing this include the analysis of clusters in periodic boundrary
    conditions and consistent center of mass calculation across box boundraries. This
    procedure results in the right center of mass as seen below ::

       +-----------+
       |           |
       x 1       2 |
       |           |
       +-----------+

    """
    theta = (ag.positions / ag.universe.dimensions[:3]) * 2 * np.pi
    xi = (np.cos(theta) * weights[:, None]).sum(axis=0) / weights.sum()
    zeta = (np.sin(theta) * weights[:, None]).sum(axis=0) / weights.sum()
    theta_com = np.arctan2(-zeta, -xi) + np.pi
    return theta_com / (2 * np.pi) * ag.universe.dimensions[:3]


def symmetrize(
    m: np.ndarray,
    axis: None | int | tuple[int] = None,
    inplace: bool = False,
    is_odd: bool = False,
) -> np.ndarray:
    """Symmeterize an array.

    The shape of the array is preserved, but the elements are symmetrized with respect
    to the given axis.

    Parameters
    ----------
    m : array_like
        Input array to symmetrize
    axis : int, tuple(int)
         Axis or axes along which to symmetrize over. The default, ``axis=None``, will
         symmetrize over all of the axes of the input array. If axis is negative it
         counts from the last to the first axis. If axis is a :obj:`tuple` of ints,
         symmetrizing is performed on all of the axes specified in the :obj:`tuple`.
    inplace : bool
        Do symmetrizations inplace. If :obj:`False` a new array is returned.
    is_odd : bool
        The parity to use for symmetrization. If :obj:`False` (default), the
        symmetrization is done with "even" parity, meaning that the output array will be
        symmetric with respect to the specified axis. If :obj:`True`, the symmetrization
        is done with "odd" parity, meaning that the output array will be antisymmetric.

    Returns
    -------
    out : array_like
        the symmetrized array

    Notes
    -----
    symmetrize uses :meth:`np.flip` for flipping the indices.

    Examples
    --------
    >>> A = np.arange(10).astype(float)
    >>> A
    array([0., 1., 2., 3., 4., 5., 6., 7., 8., 9.])
    >>> symmetrize(A)
    array([4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5])
    >>> symmetrize(A, inplace=True)
    array([4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5])
    >>> A
    array([4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5, 4.5])

    Antisymmetrization can be achieved by setting ``is_odd=True``.
    >>> A = np.arange(10).astype(float)
    >>> A
    array([0., 1., 2., 3., 4., 5., 6., 7., 8., 9.])
    >>> symmetrize(A, is_odd=True)
    array([-4.5, -3.5, -2.5, -1.5, -0.5,  0.5,  1.5,  2.5,  3.5,  4.5])

    It also works for arrays with more than 1 dimensions in a general dimension.

    >>> A = np.arange(20).astype(float).reshape(2, 10).T
    >>> A
    array([[ 0., 10.],
           [ 1., 11.],
           [ 2., 12.],
           [ 3., 13.],
           [ 4., 14.],
           [ 5., 15.],
           [ 6., 16.],
           [ 7., 17.],
           [ 8., 18.],
           [ 9., 19.]])
    >>> symmetrize(A)
    array([[9.5, 9.5],
           [9.5, 9.5],
           [9.5, 9.5],
           [9.5, 9.5],
           [9.5, 9.5],
           [9.5, 9.5],
           [9.5, 9.5],
           [9.5, 9.5],
           [9.5, 9.5],
           [9.5, 9.5]])
    >>> symmetrize(A, axis=0)
    array([[ 4.5, 14.5],
           [ 4.5, 14.5],
           [ 4.5, 14.5],
           [ 4.5, 14.5],
           [ 4.5, 14.5],
           [ 4.5, 14.5],
           [ 4.5, 14.5],
           [ 4.5, 14.5],
           [ 4.5, 14.5],
           [ 4.5, 14.5]])

    """
    # The returned array will be of type float
    out = m.copy().astype("float")
    out += (-1 if is_odd else 1) * np.flip(m, axis=axis)
    out /= 2

    if inplace:
        # To safely cast the the original array type to float in-place,
        # first change the dtype to float...
        m.dtype = np.dtype("float")
        # ...and then write the new values to the original array.
        m[...] = out
        return m
    return out


def transform_cylinder(
    positions: np.ndarray, origin: np.ndarray, dim: int
) -> np.ndarray:
    """Transform positions into cylinder coordinates.

    The origin of th coordinate system is at ``origin``, the direction of the cylinder
    is defined by ``dim``.

    Parameters
    ----------
    positions : numpy.ndarray
        Cartesian coordinates (x,y,z)
    origin : numpy.ndarray
        Origin of the new cylindrical coordinate system (x,y,z).
    dim : int
        Direction of the cylinder axis (0=x, 1=y, 2=z).

    Returns
    -------
    numpy.ndarray
        Positions in cylinder coordinates (r, phi, z)

    """
    trans_positions = np.zeros(positions.shape)

    odims = np.roll(np.arange(3), -dim)[1:]

    # shift origin to box center
    pos_xyz_center = positions - origin

    # r component
    trans_positions[:, 0] = np.linalg.norm(pos_xyz_center[:, odims], axis=1)

    # phi component
    np.arctan2(*pos_xyz_center[:, odims].T, out=trans_positions[:, 1])

    # z component
    trans_positions[:, 2] = np.copy(positions[:, dim])

    return trans_positions


def transform_sphere(positions: np.ndarray, origin: np.ndarray) -> np.ndarray:
    """Transform positions into spherical coordinates.

    The origin of the new coordinate system is at ``origin``.

    Parameters
    ----------
    positions : numpy.ndarray
        Cartesian coordinates ``(x,y,z)``
    origin : numpy.ndarray
        Origin of the new spherical coordinate system ``(x,y,z)``.

    Returns
    -------
    numpy.ndarray
        Positions in spherical coordinates (:math:`r`, phi, theta)

    """
    trans_positions = np.zeros(positions.shape)

    # shift origin to box center
    # positions -= origin
    pos_xyz_center = positions - origin

    # r component
    trans_positions[:, 0] = np.linalg.norm(pos_xyz_center, axis=1)
    # phi component
    np.arctan2(pos_xyz_center[:, 1], pos_xyz_center[:, 0], out=trans_positions[:, 1])
    # theta component
    np.arccos(pos_xyz_center[:, 2] / trans_positions[:, 0], out=trans_positions[:, 2])

    return trans_positions


def combine_subsample_variance(n_A, n_B, mu_A, mu_B, M_A, M_B):
    """Calculate the mean and variance of two datasets of arbitrary size.

    Given two datasets of arbitrary size, this function calculates the mean and
    variance of the combined dataset using Welford's method. This is useful for
    parallel processing where the datasets are processed in chunks.

    Parameters
    ----------
    n_A : int
        Number of elements in dataset A.
    n_B : int
        Number of elements in dataset B.
    mu_A : float
        Mean of dataset A.
    mu_B : float
        Mean of dataset B.
    M_A : float
        Sum of squares of deviations from the mean for dataset A.
    M_B : float
        Sum of squares of deviations from the mean for dataset B.

    Returns
    -------
    n_AB : int
        Total number of elements in the combined dataset (n_A + n_B).
    mu_AB : float
        Mean of the combined dataset.
    M_AB : float
        Sum of squares of deviations from the mean for the combined dataset.
    """
    n_AB = n_A + n_B
    delta = np.nan_to_num(mu_B) - np.nan_to_num(mu_A)
    mu_AB = np.nan_to_num(mu_A) + delta * n_B / n_AB
    M_AB = np.nan_to_num(M_A) + np.nan_to_num(M_B) + delta**2 * n_A * n_B / n_AB

    return n_AB, mu_AB, M_AB



#!/usr/bin/env python
#
# Copyright (c) 2026 Authors and contributors
# (see the AUTHORS.rst file for the full list of names)
#
# Released under the GNU Public Licence, v3 or any higher version
# SPDX-License-Identifier: GPL-3.0-or-later
"""Module for computing kinetic energy timeseries."""

import MDAnalysis as mda


class KineticEnergy:
    """Class has been removed and is no longer available."""

    def __init__(
        self,
        atomgroup: mda.AtomGroup,  # noqa
        refpoint: str = "com",  # noqa
        refgroup: mda.AtomGroup | None = None,  # noqa
        unwrap: bool = False,  # noqa
        pack: bool = True,  # noqa
        jitter: float = 0.0,  # noqa
        concfreq: int = 0,  # noqa
        output: str = "ke.dat",  # noqa
    ) -> None:
        raise ValueError(self.__doc__)



#!/usr/bin/env python
#
# Copyright (c) 2026 Authors and contributors
# (see the AUTHORS.rst file for the full list of names)
#
# Released under the GNU Public Licence, v3 or any higher version
# SPDX-License-Identifier: GPL-3.0-or-later
"""Module for computing dielectric spectra for bulk systems."""

import MDAnalysis as mda


class DielectricSpectrum:
    """Class has been moved to the ``spectrakit`` package."""

    def __init__(
        self,
        atomgroup: mda.AtomGroup,  # noqa
        temperature: float = 300,  # noqa
        segs: int = 20,  # noqa
        df: float | None = None,  # noqa
        bins: int = 200,  # noqa
        binafter: float = 20,  # noqa
        nobin: bool = False,  # noqa
        refgroup: mda.AtomGroup | None = None,  # noqa
        unwrap: bool = True,  # noqa
        pack: bool = True,  # noqa
        jitter: float = 0.0,  # noqa
        concfreq: int = 0,  # noqa
        output_prefix: str = "",  # noqa
    ) -> None:
        raise ValueError(self.__doc__)



#!/usr/bin/env python
#
# Copyright (c) 2026 Authors and contributors
# (see the AUTHORS.rst file for the full list of names)
#
# Released under the GNU Public Licence, v3 or any higher version
# SPDX-License-Identifier: GPL-3.0-or-later
"""Analyse molecular dynamics simulation of interfacial and confined systems."""

from mdacli import cli

from maicos import __version__
from maicos.core import AnalysisBase


def main():
    """Execute main CLI entry point."""
    cli(
        name="MAICoS",
        module_list=["maicos"],
        base_class=AnalysisBase,
        version=__version__,
        description=__doc__,
        ignore_warnings=True,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python
#
# Copyright (c) 2026 Authors and contributors
# (see the AUTHORS.rst file for the full list of names)
#
# Released under the GNU Public Licence, v3 or any higher version
# SPDX-License-Identifier: GPL-3.0-or-later
r"""Library modules of MAICoS.

This library contains additional modules, like general and mathematical helper
functions, which are used in the other MAICoS modules.
"""

__all__ = [
    "util",
    "math",
    "weights",
]
